# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Simple Java Classfile unpacking module. Can be made to act an
awful lot like the javap utility included with most Java SDKs.

Most of the information used to write this was gathered from the
following web pages

References
----------
* http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html
* http://java.sun.com/docs/books/jvms/second_edition/html/VMSpecTOC.doc.html
* http://en.wikipedia.org/wiki/Class_(file_format)

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""  # noqa


from functools import partial
from six.moves import range

from .dirutils import fnmatches
from .opcodes import disassemble
from .pack import compile_struct, unpack, UnpackException

try:
    buffer
except NameError:
    buffer = memoryview


__all__ = (
    "JavaClassInfo", "JavaConstantPool", "JavaMemberInfo",
    "JavaCodeInfo", "JavaExceptionInfo", "JavaInnerClassInfo",
    "JavaAnnotation",
    "NoPoolException", "Unimplemented", "ClassUnpackException",
    "UnknownConstantPoolTagException", "UnpackException",
    "platform_from_version",
    "is_class", "is_class_file",
    "unpack_class", "unpack_classfile",
    "CONST_Utf8", "CONST_Integer", "CONST_Float",
    "CONST_Long", "CONST_Double", "CONST_Class",
    "CONST_String", "CONST_Fieldref", "CONST_Methodref",
    "CONST_InterfaceMethodref", "CONST_NameAndType",
    "CONST_ModuleId", "CONST_MethodHandle",
    "CONST_MethodType", "CONST_InvokeDynamic",
    "CONST_Module", "CONST_Package",
    "ACC_PUBLIC", "ACC_PRIVATE", "ACC_PROTECTED",
    "ACC_STATIC", "ACC_FINAL", "ACC_SYNCHRONIZED",
    "ACC_SUPER", "ACC_VOLATILE", "ACC_BRIDGE",
    "ACC_TRANSIENT", "ACC_VARARGS", "ACC_NATIVE",
    "ACC_INTERFACE", "ACC_ABSTRACT", "ACC_STRICT",
    "ACC_SYNTHETIC", "ACC_ANNOTATION", "ACC_ENUM",
    "ACC_MODULE",
)


# the four bytes at the start of every class file
JAVA_CLASS_MAGIC = (0xCA, 0xFE, 0xBA, 0xBE)


_BUFFERING = 2 ** 14


# The constant pool types
# pylint: disable=C0103
CONST_Utf8 = 1
CONST_Integer = 3
CONST_Float = 4
CONST_Long = 5
CONST_Double = 6
CONST_Class = 7
CONST_String = 8
CONST_Fieldref = 9
CONST_Methodref = 10
CONST_InterfaceMethodref = 11
CONST_NameAndType = 12
CONST_ModuleId = 13  # Removed? Maybe OpenJDK only?
CONST_MethodHandle = 15
CONST_MethodType = 16
CONST_Dynamic = 17
CONST_InvokeDynamic = 18
CONST_Module = 19
CONST_Package = 20


# class and member flags
ACC_PUBLIC = 0x0001
ACC_PRIVATE = 0x0002
ACC_PROTECTED = 0x0004
ACC_STATIC = 0x0008
ACC_FINAL = 0x0010
ACC_SYNCHRONIZED = 0x0020
ACC_SUPER = 0x0020
ACC_VOLATILE = 0x0040
ACC_BRIDGE = 0x0040
ACC_TRANSIENT = 0x0080
ACC_VARARGS = 0x0080
ACC_NATIVE = 0x0100
ACC_INTERFACE = 0x0200
ACC_ABSTRACT = 0x0400
ACC_STRICT = 0x0800
ACC_SYNTHETIC = 0x1000
ACC_ANNOTATION = 0x2000
ACC_ENUM = 0x4000
ACC_MODULE = 0x8000


# commonly re-occurring struct formats
_B = compile_struct(">B")
_BBBB = compile_struct(">BBBB")
_BH = compile_struct(">BH")
_H = compile_struct(">H")
_HH = compile_struct(">HH")
_HHH = compile_struct(">HHH")
_HHHH = compile_struct(">HHHH")
_HHHHH = compile_struct(">HHHHH")
_HI = compile_struct(">HI")
_HHI = compile_struct(">HHI")


class NoPoolException(Exception):
    """
    raised by methods that need a JavaConstantPool, but aren't
    provided one on the owning instance
    """

    pass


class Unimplemented(Exception):
    """
    raised when something unexpected happens, which usually indicates
    part of the classfile specification that wasn't implemented in
    this module yet
    """

    pass


class ClassUnpackException(Exception):
    """
    raised when a class couldn't be unpacked
    """

    pass


class UnknownConstantPoolTagException(Exception):
    """
    raised when a constant pool tag is unknown and probably corrupted
    """

    pass


class JavaConstantPool(object):
    """
    A constants pool

    reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.4
    """  # noqa

    def __init__(self):
        self.consts = tuple()


    def __eq__(self, other):
        return (isinstance(other, JavaConstantPool) and
                (self.consts == other.consts))


    def __ne__(self, other):
        return not self.__eq__(other)


    def unpack(self, unpacker):
        """
        Unpacks the constant pool from an unpacker stream
        """

        (count, ) = unpacker.unpack_struct(_H)

        # first item is never present in the actual data buffer, but
        # the count number acts like it would be.
        items = [(None, None), ]
        count -= 1

        # Long and Double const types will "consume" an item count,
        # but not data
        hackpass = False

        for _i in range(0, count):

            if hackpass:
                # previous item was a long or double
                hackpass = False
                items.append((None, None))

            else:
                item = _unpack_const_item(unpacker)
                items.append(item)

                # if this item was a long or double, skip the next
                # counter.
                if item[0] in (CONST_Long, CONST_Double):
                    hackpass = True

        self.consts = items


    def get_const(self, index):
        """
        returns the type and value of the constant at index
        """

        return self.consts[index]


    def deref_const(self, index):
        """
        returns the dereferenced value from the const pool. For simple
        types, this will be a single value indicating the constant.
        For more complex types, such as fieldref, methodref, etc, this
        will return a tuple.
        """

        if not index:
            raise IndexError("Requested const 0")

        t, v = self.consts[index]

        # CONSTANT_info {
        #     u1 tag;
        #     u4 bytes; (the value of the constant)
        # }
        # NOTE: Long and Double has high_bytes and low_bytes.
        # NOTE: Utf8 has two fields length and bytes
        if t in (CONST_Utf8, CONST_Integer, CONST_Float,
                 CONST_Long, CONST_Double):
            return v

        # CONSTANT_info {
        #     u1 tag;
        #     u2 index; (valid index into the constant_pool)
        # }
        # NOTE: each constant can have a little bit different field name
        elif t in (CONST_Class, CONST_String, CONST_MethodType,
                   CONST_Module, CONST_Package):
            return self.deref_const(v)

        # CONSTANT_info {
        #     u1 tag;
        #     u2 index; (valid index into the constant_pool)
        #     u2 additional_index; (valid index into the constant_pool)
        # }
        # NOTE: each constant can have a little bit different field name
        elif t in (CONST_Fieldref, CONST_Methodref,
                   CONST_InterfaceMethodref, CONST_NameAndType,
                   CONST_ModuleId):
            return tuple(self.deref_const(i) for i in v)

        # CONSTANT_info {
        #     u1 tag;
        #     u2 bootstrap_method_attr_index; (must be a valid index into the
        #                                      bootstrap_methods array
        #                                      of the bootstrap method table of
        #                                      this class file)
        #     u2 name_and_type_index; (valid index into the constant_pool)
        # }
        elif t in (CONST_InvokeDynamic, CONST_Dynamic):
            # TODO: v[0] needs to come from the bootstrap methods table
            return (v[0], self.deref_const(v[1]))

        # CONSTANT_info {
        #     u1 tag;
        #     u1 reference_kind; (must be in the range 1 to 9)
        #     u2 reference_index; (valid index into the constant_pool)
        # }
        elif t == CONST_MethodHandle:
            # TODO: v[0] treat index according to its kind
            return (v[0], self.deref_const(v[1]))

        else:
            raise UnknownConstantPoolTagException(
                "Unknown constant pool type %r" % t)


    def constants(self):
        """
        sequence of tuples (index, type, dereferenced value) of the
        constant pool entries.
        """

        for i in range(1, len(self.consts)):
            t, _v = self.consts[i]
            if t:
                yield (i, t, self.deref_const(i))


    def pretty_constants(self):
        """
        the sequence of tuples (index, pretty type, value) of the constant
        pool entries.
        """

        for i in range(1, len(self.consts)):
            t, v = self.pretty_const(i)
            if t:
                yield (i, t, v)


    def pretty_const(self, index):
        """
        a tuple of the pretty type and val, or (None, None) for invalid
        indexes (such as the second part of a long or double value)
        """

        t, v = self.consts[index]
        if not t:
            return None, None
        else:
            return _pretty_const_type_val(t, v)


    def pretty_deref_const(self, index):
        """
        A string representation of the end-value of a constant.  This will
        deref the constant index, and if it is a compound type, will
        continue dereferencing until it can compose the full value
        (eg: a CONST_Methodref will be composed of its class, name,
        and value derefenced constants)
        """

        t, v = self.consts[index]

        if t == CONST_String:
            result = self.deref_const(v)

        elif t in (CONST_Utf8,
                   CONST_Integer, CONST_Float,
                   CONST_Long, CONST_Double):
            result = v

        elif t == CONST_Class:
            result = _pretty_class(self.deref_const(v))

        elif t == CONST_Fieldref:
            cn = self.deref_const(v[0])
            cn = _pretty_class(cn)

            n, t = self.deref_const(v[1])

            result = "%s.%s:%s" % (cn, n, _pretty_type(t))

        elif t in (CONST_Methodref,
                   CONST_InterfaceMethodref):

            cn = self.deref_const(v[0])
            cn = _pretty_class(cn)

            n, t = self.deref_const(v[1])

            args, ret = tuple(_pretty_typeseq(t))

            result = "%s.%s%s:%s" % (cn, n, args, ret)

        elif t == CONST_NameAndType:
            a, b = (self.deref_const(i) for i in v)
            b = "".join(_pretty_typeseq(b))
            result = "%s:%s" % (a, b)

        elif t == CONST_ModuleId:
            a, b = (self.deref_const(i) for i in v)
            result = "%s@%s" % (a, b)

        elif t == CONST_InvokeDynamic:
            # TODO: v[0] needs to come from the bootstrap methods table
            result = "InvokeDynamic %r %r" % (v[0], self.deref_const(v[1]))

        elif t == CONST_Dynamic:
            # TODO: v[0] needs to come from the bootstrap methods table
            result = "Dynamic %r %r" % (v[0], self.deref_const(v[1]))

        elif t == CONST_MethodType:
            result = "MethodType %r %r" % (v[0], self.deref_const(v[1]))

        elif t == CONST_Module:
            result = "Module %s" % self.deref_cons(v[0])

        elif t == CONST_Package:
            result = "Package %s" % self.deref_cons(v[0])

        elif not t:
            # the skipped-type, meaning the prior index was a
            # two-slotter.
            result = ""

        else:
            raise UnknownConstantPoolTagException(
                "No pretty for const type %r" % t)

        return result


class JavaAttributes(dict):
    """
    attributes table, as used in class, member, and code
    structures. Requires access to a JavaConstantPool instance for
    many of its methods to work correctly.
    """

    def __init__(self, cpool):
        dict.__init__(self)
        self.cpool = cpool


    def unpack(self, unpacker):
        """
        Unpack an attributes table from an unpacker stream.  Modifies the
        structure of this instance.
        """

        # bound method for dereferencing constants
        cval = self.cpool.deref_const

        (count,) = unpacker.unpack_struct(_H)
        for _i in range(0, count):
            (name, size) = unpacker.unpack_struct(_HI)
            self[cval(name)] = unpacker.read(size)


class JavaClassInfo(object):
    """
    Information from a disassembled Java class file.

    reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html
    """

    def __init__(self):
        self.cpool = JavaConstantPool()
        self.attribs = JavaAttributes(self.cpool)

        self.magic = JAVA_CLASS_MAGIC
        self.version = (0, 0)
        self.access_flags = 0
        self.this_ref = 0
        self.super_ref = 0
        self.interfaces = tuple()
        self.fields = tuple()
        self.methods = tuple()
        self.annotations = None
        self.invisible_annotations = None

        self._provides = None
        self._provides_private = None
        self._requires = None


    def deref_const(self, index):
        """
        dereference a value from the parent constant pool
        """

        return self.cpool.deref_const(index)


    def get_attribute(self, name):
        """
        get an attribute buffer by name
        """

        return self.attribs.get(name)


    def unpack(self, unpacker, magic=None):
        """
        Unpacks a Java class from an unpacker stream. Updates the
        structure of this instance.

        If the unpacker has already had the magic header read off of
        it, the read value may be passed via the optional magic
        parameter and it will not attempt to read the value again.
        """

        # only unpack the magic bytes if it wasn't specified
        magic = magic or unpacker.unpack_struct(_BBBB)

        if isinstance(magic, (str, buffer)):
            magic = tuple(ord(m) for m in magic)
        else:
            magic = tuple(magic)

        if magic != JAVA_CLASS_MAGIC:
            raise ClassUnpackException("Not a Java class file")

        self.magic = magic

        # unpack (minor, major), store as (major, minor)
        self.version = unpacker.unpack_struct(_HH)[::-1]

        # unpack constant pool
        self.cpool.unpack(unpacker)

        (a, b, c) = unpacker.unpack_struct(_HHH)
        self.access_flags = a
        self.this_ref = b
        self.super_ref = c

        # unpack interfaces
        (count,) = unpacker.unpack_struct(_H)
        self.interfaces = unpacker.unpack(">%iH" % count)

        uobjs = unpacker.unpack_objects

        # unpack fields
        self.fields = tuple(uobjs(JavaMemberInfo,
                                  self.cpool, is_method=False))

        # unpack methods
        self.methods = tuple(uobjs(JavaMemberInfo,
                                   self.cpool, is_method=True))

        # unpack attributes
        self.attribs.unpack(unpacker)


    def get_field_by_name(self, name):
        """
        the field member matching name, or None if no such field is found
        """

        for f in self.fields:
            if f.get_name() == name:
                return f
        return None


    def get_methods_by_name(self, name):
        """
        generator of methods matching name. This will include any bridges
        present.
        """

        return (m for m in self.methods if m.get_name() == name)


    def get_method(self, name, arg_types=()):
        """
        searches for the method matching the name and having argument type
        descriptors matching those in arg_types.

        Parameters
        ==========
        arg_types : sequence of strings
          each string is a parameter type, in the non-pretty format.

        Returns
        =======
        method : `JavaMemberInfo` or `None`
          the single matching, non-bridging method of matching name
          and parameter types.
        """

        # ensure any lists or iterables are converted to tuple for
        # comparison against get_arg_type_descriptors()
        arg_types = tuple(arg_types)

        for m in self.get_methods_by_name(name):
            if (((not m.is_bridge()) and
                 m.get_arg_type_descriptors() == arg_types)):
                return m
        return None


    def get_method_bridges(self, name, arg_types=()):
        """
        generator of bridge methods found that adapt the return types of a
        named method and having argument type descriptors matching
        those in arg_types.
        """

        for m in self.get_methods_by_name(name):
            if ((m.is_bridge() and
                 m.get_arg_type_descriptors() == arg_types)):
                yield m


    def get_version(self):
        """
        the (major, minor) version of Java required by this Java class
        """

        return self.version


    def get_platform(self):
        """
        The platform as a string, derived from the major and minor version
        number
        """

        return platform_from_version(*self.version)


    def is_public(self):
        """
        is this class public
        """

        return self.access_flags & ACC_PUBLIC


    def is_final(self):
        """
        is this class final
        """

        return self.access_flags & ACC_FINAL


    def is_super(self):
        """
        class has the Super flag set.

        This flag is used by the JVM to differentiate the behavior in
        the method resolution order of the class.
        """

        return self.access_flags & ACC_SUPER


    def is_interface(self):
        """
        is this an interface
        """

        return self.access_flags & ACC_INTERFACE


    def is_abstract(self):
        """
        is this an abstract class
        """

        return self.access_flags & ACC_ABSTRACT


    def is_annotation(self):
        """
        is this an annotation class
        """

        return self.access_flags & ACC_ANNOTATION


    def is_enum(self):
        """
        is this an enum class
        """

        return self.access_flags & ACC_ENUM


    def get_this(self):
        """
        the name of this class
        """

        return self.deref_const(self.this_ref)


    def is_deprecated(self):
        """
        is this class deprecated

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.15
        """  # noqa

        return bool(self.get_attribute("Deprecated"))


    def get_super(self):
        """
        get the parent class that this extends
        """

        return self.deref_const(self.super_ref) if self.super_ref else ""


    def get_interfaces(self):
        """
        tuple of interfaces that this class implements
        """

        return tuple(self.deref_const(i) for i in self.interfaces)


    def _get_annotations(self, python_attr_name, java_attr_name):

        annos = getattr(self, python_attr_name)

        if annos is None:
            buff = self.get_attribute(java_attr_name)
            if buff is None:
                annos = tuple()

            else:
                with unpack(buff) as up:
                    annos = up.unpack_objects(JavaAnnotation, self.cpool)
                    annos = tuple(annos)

            setattr(self, python_attr_name, annos)

        return annos


    def get_annotations(self):
        """
        The RuntimeVisibleAnnotations attribute. A tuple of JavaAnnotation
        instances

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.16
        """  # noqa

        return self._get_annotations("annotations",
                                     "RuntimeVisibleAnnotations")


    def get_invisible_annotations(self):
        """
        The RuntimeInvisibleAnnotations attribute. A tuple of
        JavaAnnotation instances

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.17
        """  # noqa

        return self._get_annotations("invisible_annotations",
                                     "RuntimeInvisibleAnnotations")


    def get_sourcefile(self):
        """
        the name of thie file this class was compiled from, or None if not
        indicated

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.10
        """  # noqa

        buff = self.get_attribute("SourceFile")
        if buff is None:
            return None

        with unpack(buff) as up:
            (ref,) = up.unpack_struct(_H)

        return self.deref_const(ref)


    def get_source_debug_extension(self):
        """
        reference:
        http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.11
        """  # noqa

        buff = self.get_attribute("SourceDebugExtension")
        return (buff and str(buff)) or None


    def get_innerclasses(self):
        """
        sequence of JavaInnerClassInfo instances describing the inner
        classes of this class definition

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.6
        """  # noqa

        buff = self.get_attribute("InnerClasses")
        if buff is None:
            return tuple()

        with unpack(buff) as up:
            return tuple(up.unpack_objects(JavaInnerClassInfo, self.cpool))


    def get_signature(self):
        """
        the generics class signature

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.9
        """  # noqa

        buff = self.get_attribute("Signature")
        if buff is None:
            return None

        with unpack(buff) as up:
            (ref,) = up.unpack_struct(_H)

        return self.deref_const(ref)


    def pretty_signature(self):
        """
        pretty version of the signature
        """

        return pretty_generic(self.get_signature())


    def get_enclosingmethod(self):
        """
        the class.method or class (if the definition is not from within a
        method) that encloses the definition of this class. Returns
        None if this was not an inner class.

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.7
        """  # noqa

        buff = self.get_attribute("EnclosingMethod")

        # TODO:
        # Running across classes with data in this attribute like
        # 00 06 00 00
        # which would be the 6th const for the class name, and the
        # zero-th (INVALID) const for method. Maybe this is static
        # inner classes?

        if buff is None:
            return None

        # class index, method index
        with unpack(buff) as up:
            ci, mi = up.unpack_struct(_HH)

        result = None

        if ci and mi:
            enc_class = self.deref_const(ci)
            enc_meth, enc_type = self.deref_const(mi)
            result = "%s.%s%s" % (enc_class, enc_meth, enc_type)

        elif ci:
            result = self.deref_const(ci)

        return result


    def _pretty_access_flags_gen(self):
        """
        generator of the pretty access flags
        """

        if self.is_public():
            yield "public"
        if self.is_final():
            yield "final"
        if self.is_abstract():
            yield "abstract"
        if self.is_interface():
            if self.is_annotation():
                yield "@interface"
            else:
                yield "interface"
        if self.is_enum():
            yield "enum"


    def pretty_access_flags(self):
        """
        generator of the pretty access flag names
        """

        return self._pretty_access_flags_gen()


    def pretty_this(self):
        """
        the pretty version of this class name
        """

        return _pretty_class(self.get_this())


    def pretty_super(self):
        """
        the pretty version of the parent class name
        """

        return _pretty_class(self.get_super())


    def pretty_interfaces(self):
        """
        the pretty versions of any interfaces this class implements
        """

        return (_pretty_class(t) for t in self.get_interfaces())


    def pretty_descriptor(self):
        """
        get the class or interface name, its accessor flags, its parent
        class, and any interfaces it implements
        """

        f = " ".join(self.pretty_access_flags())
        if not self.is_interface():
            f += " class"

        n = self.pretty_this()
        e = self.pretty_super()
        i = ",".join(self.pretty_interfaces())

        if i:
            return "%s %s extends %s implements %s" % (f, n, e, i)
        else:
            return "%s %s extends %s" % (f, n, e)


    def _get_provides(self, private=False):
        """
        iterator of provided classes, fields, methods
        """

        # TODO I probably need to add inner classes here

        me = self.pretty_this()
        yield me

        for field in self.fields:
            if private or field.is_public():
                yield "%s.%s" % (me, field.pretty_identifier())

        for method in self.methods:
            if private or method.is_public():
                yield "%s.%s" % (me, method.pretty_identifier())


    def _get_requires(self):
        """
        iterator of required classes, fields, methods, determined my
        mining the constant pool for such types
        """

        provided = set(self.get_provides(private=True))
        cpool = self.cpool

        # loop through the constant pool for API types
        for i, t, _v in cpool.constants():

            if t in (CONST_Class, CONST_Fieldref,
                     CONST_Methodref, CONST_InterfaceMethodref):

                # convert this away from unicode so we can
                pv = str(cpool.pretty_deref_const(i))

                if pv[0] == "[":
                    # sometimes when calling operations on an array
                    # the type embeded in the cpool will be the array
                    # type, not just the class type. Let's only gather
                    # the types themselves and ignore the fact that
                    # the class really wanted an array of them.  In
                    # the event that this was a method or field on the
                    # array, we'll throw away that as well, and just
                    # emit the type contained in the array.
                    t = _typeseq(pv)
                    if t[1] == "L":
                        pv = _pretty_type(t[1:])
                    else:
                        pv = None

                if pv and (pv not in provided):
                    yield pv


    def get_provides(self, ignored=tuple(), private=False):
        """
        The provided API, including the class itself, its fields, and its
        methods.
        """

        if private:
            if self._provides_private is None:
                self._provides_private = set(self._get_provides(True))
            provides = self._provides_private
        else:
            if self._provides is None:
                self._provides = set(self._get_provides(False))
            provides = self._provides

        return [prov for prov in provides if not fnmatches(prov, *ignored)]


    def get_requires(self, ignored=tuple()):
        """
        The required API, including all external classes, fields, and
        methods that this class references
        """

        if self._requires is None:
            self._requires = set(self._get_requires())

        requires = self._requires
        return [req for req in requires if not fnmatches(req, *ignored)]


class JavaMemberInfo(object):
    """
    A field or method of a java class
    """

    def __init__(self, cpool, is_method=False):
        self.cpool = cpool
        self.attribs = JavaAttributes(cpool)
        self.access_flags = 0
        self.name_ref = 0
        self.descriptor_ref = 0
        self.is_method = is_method
        self.annotations = None
        self.invisible_annotations = None
        self.parameter_annotations = None
        self.invisible_parameter_annotations = None


    def deref_const(self, index):
        """
        Dereference a constant in the parent constant pool
        """

        return self.cpool.deref_const(index)


    def get_attribute(self, name):
        """
        Get an attribute buffer by name
        """

        return self.attribs.get(name)


    def get_signature(self):
        """
        the Signature attribute
        """

        buff = self.get_attribute("Signature")
        if buff is None:
            return None

        # type index
        with unpack(buff) as up:
            (ti,) = up.unpack_struct(_H)

        return self.deref_const(ti)


    def pretty_signature(self):
        """
        pretty version of the signature
        """

        return pretty_generic(self.get_signature())


    def get_module(self):
        """
        the Module attribute
        """

        buff = self.get_attribute("Module")
        if buff is None:
            return None

        with unpack(buff) as up:
            (ti,) = up.unpack_struct(_H)

        return self.deref_const(ti)


    def unpack(self, unpacker):
        """
        unpack the contents of this instance from the values in unpacker
        """

        (a, b, c) = unpacker.unpack_struct(_HHH)

        self.access_flags = a
        self.name_ref = b
        self.descriptor_ref = c
        self.attribs.unpack(unpacker)


    def get_name(self):
        """
        the name of this member
        """

        return self.deref_const(self.name_ref)


    def get_descriptor(self):
        """
        the descriptor of this member
        """

        return self.deref_const(self.descriptor_ref)


    def is_public(self):
        """
        is this member public
        """

        return self.access_flags & ACC_PUBLIC


    def is_private(self):
        """
        is this member private
        """

        return self.access_flags & ACC_PRIVATE


    def is_protected(self):
        """
        is this member protected
        """

        return self.access_flags & ACC_PROTECTED


    def is_static(self):
        """
        is this member static
        """

        return self.access_flags & ACC_STATIC


    def is_final(self):
        """
        is this member final
        """

        return self.access_flags & ACC_FINAL


    def is_synchronized(self):
        """
        is this member synchronized
        """

        return self.access_flags & ACC_SYNCHRONIZED


    def is_native(self):
        """
        is this member native
        """

        return self.access_flags & ACC_NATIVE


    def is_abstract(self):
        """
        is this member abstract
        """

        return self.access_flags & ACC_ABSTRACT


    def is_strict(self):
        """
        is this member strict
        """

        return self.access_flags & ACC_STRICT


    def is_volatile(self):
        """
        is this member volatile
        """

        return self.access_flags & ACC_VOLATILE


    def is_transient(self):
        """
        is this member transient
        """

        return self.access_flags & ACC_TRANSIENT


    def is_bridge(self):
        """
        is this method a bridge to another method
        """

        return self.access_flags & ACC_BRIDGE


    def is_varargs(self):
        """
        is this a varargs method
        """

        return self.access_flags & ACC_VARARGS


    def is_synthetic(self):
        """
        is this a synthetic method

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.8
        """  # noqa

        return ((self.access_flags & ACC_SYNTHETIC) or
                bool(self.get_attribute("Synthetic")))


    def is_enum(self):
        """
        it this member an enum
        """

        return self.access_flags & ACC_ENUM


    def is_module(self):
        """
        is this a module member
        """

        return self.access_flags & ACC_MODULE


    def is_deprecated(self):
        """
        is this member deprecated

        reference: http://docs.oracle.com/javase/specs/jvms/se5.0/html/ClassFile.doc.html#78232
        """  # noqa

        return bool(self.get_attribute("Deprecated"))


    def _get_annotations(self, python_attr_name, java_attr_name,
                         for_params=False):

        annos = getattr(self, python_attr_name)

        if annos is None:
            buff = self.get_attribute(java_attr_name)
            if buff is None:
                annos = tuple()

            else:
                with unpack(buff) as up:
                    unp = partial(up.unpack_objects,
                                  JavaAnnotation, self.cpool)

                    if for_params:
                        (param_count, ) = up.unpack_struct(_B)
                        annos = (tuple(unp()) for _i in range(param_count))
                    else:
                        annos = unp()
                    annos = tuple(annos)

            setattr(self, python_attr_name, annos)

        return annos


    def get_annotations(self):
        """
        The RuntimeVisibleAnnotations attribute. A tuple of JavaAnnotation
        instances

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.16
        """  # noqa

        return self._get_annotations("annotations",
                                     "RuntimeVisibleAnnotations")


    def get_invisible_annotations(self):
        """
        The RuntimeInvisibleAnnotations attribute. A tuple of
        JavaAnnotation instances

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.17
        """  # noqa

        return self._get_annotations("invisible_annotations",
                                     "RuntimeInvisibleAnnotations")


    def get_parameter_annotations(self):
        """
        The RuntimeVisibleParameterAnnotations attribute.  Contains a
        tuple of JavaAnnotation instances for each param.

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.18
        """  # noqa
        return self._get_annotations("parameter_annotations",
                                     "RuntimeVisibleParameterAnnotations",
                                     for_params=True)


    def get_invisible_parameter_annotations(self):
        """
        The RuntimeInvisibleParameterAnnotations attribute.  Contains a
        tuple of JavaAnnotation instances for each param.

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.19
        """  # noqa
        return self._get_annotations("invisible_parameter_annotations",
                                     "RuntimeInvisibleParameterAnnotations",
                                     for_params=True)


    def get_annotationdefault(self):
        """
        The AnnotationDefault attribute, only present upon fields in an
        annotaion.

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.20
        """  # noqa

        buff = self.get_attribute("AnnotationDefault")
        if buff is None:
            return None

        with unpack(buff) as up:
            (ti, ) = up.unpack_struct(_H)

        return ti


    def deref_annotationdefault(self):
        """
        dereferences the AnnotationDefault attribute
        """

        index = self.get_annotationdefault()
        if index is None:
            return None
        else:
            return self.deref_const(index)


    def get_code(self):
        """
        the JavaCodeInfo of this member if it is a non-abstract method,
        None otherwise

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.3
        """  # noqa

        buff = self.get_attribute("Code")
        if buff is None:
            return None

        with unpack(buff) as up:
            code = JavaCodeInfo(self.cpool)
            code.unpack(up)

        return code


    def get_exceptions(self):
        """
        a tuple of class names for the exception types this method may
        raise, or None if this is not a method

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.5
        """  # noqa

        buff = self.get_attribute("Exceptions")
        if buff is None:
            return ()

        with unpack(buff) as up:
            return tuple(self.deref_const(e[0]) for e
                         in up.unpack_struct_array(_H))


    def get_constantvalue(self):
        """
        the constant pool index for this field, or None if this is not a
        contant field

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.2
        """  # noqa

        buff = self.get_attribute("ConstantValue")
        if buff is None:
            return None

        with unpack(buff) as up:
            (cval_ref, ) = up.unpack_struct(_H)

        return cval_ref


    def deref_constantvalue(self):
        """
        the value in the constant pool at the get_constantvalue() index
        """

        index = self.get_constantvalue()
        if index is None:
            return None
        else:
            return self.deref_const(index)


    def get_type_descriptor(self):
        """
        the type descriptor for a field, or the return type descriptor for
        a method. Type descriptors are shorthand identifiers for the
        builtin java types.
        """

        return _typeseq(self.get_descriptor())[-1]


    def get_arg_type_descriptors(self):
        """
        The parameter type descriptor list for a method, or None for a
        field.  Type descriptors are shorthand identifiers for the
        builtin java types.
        """

        if not self.is_method:
            return tuple()

        desc = self.get_descriptor()

        tp = _typeseq(desc)
        tp = _typeseq(tp[0][1:-1])

        return tp


    def pretty_type(self):
        """
        The pretty version of get_type_descriptor.
        """

        return _pretty_type(self.get_type_descriptor())


    def pretty_arg_types(self):
        """
        Sequence of pretty argument types.
        """

        if self.is_method:
            types = self.get_arg_type_descriptors()
            return (_pretty_type(t) for t in types)
        else:
            return tuple()


    def pretty_descriptor(self):
        """
        assemble a long member name from access flags, type, argument
        types, exceptions as applicable
        """

        f = " ".join(self.pretty_access_flags())
        p = self.pretty_type()
        n = self.get_name()
        t = ",".join(self.pretty_exceptions())

        if n == "<init>":
            # we pretend that there's no return type, even though it's
            # V for constructors
            p = None

        if self.is_method:
            # stick the name and args together so there's no space
            n = "%s(%s)" % (n, ",".join(self.pretty_arg_types()))

        if t:
            # assemble any throws as necessary
            t = "throws " + t

        return " ".join(z for z in (f, p, n, t) if z)


    def _pretty_access_flags_gen(self, showall=False):

        if self.is_public():
            yield "public"
        if self.is_private():
            yield "private"
        if self.is_protected():
            yield "protected"
        if self.is_static():
            yield "static"
        if self.is_final():
            yield "final"
        if self.is_strict():
            yield "strict"
        if self.is_native():
            yield "native"
        if self.is_abstract():
            yield "abstract"
        if self.is_enum():
            yield "enum"
        if self.is_module():
            yield "module"

        if showall and self.is_synthetic():
            yield "synthetic"

        if self.is_method:
            if self.is_synchronized():
                yield "synchronized"

            if showall and self.is_bridge():
                yield "bridge"
            if showall and self.is_varargs():
                yield "varargs"

        else:
            if self.is_transient():
                yield "transient"
            if self.is_volatile():
                yield "volatile"


    def pretty_access_flags(self, showall=False):
        """
        generator of the keywords determined from the access flags
        """

        return self._pretty_access_flags_gen(showall)


    def pretty_exceptions(self):
        """
        sequence of pretty names for get_exceptions()
        """

        return (_pretty_class(e) for e in self.get_exceptions())


    def get_identifier(self):
        """
        For methods this is the return type, the name and the (non-pretty)
        argument descriptor. For fields it is simply the name.

        The return-type of methods is attached to the identifier when
        it is a bridge method, which can technically allow two methods
        with the same name and argument type list, but with different
        return type.
        """

        ident = self.get_name()

        if self.is_method:
            args = ",".join(self.get_arg_type_descriptors())
            if self.is_bridge():
                ident = "%s(%s):%s" % (ident, args, self.get_descriptor())
            else:
                ident = "%s(%s)" % (ident, args)

        return ident


    def pretty_identifier(self):
        """
        The pretty version of get_identifier
        """

        ident = self.get_name()
        if self.is_method:
            args = ",".join(self.pretty_arg_types())
            ident = "%s(%s)" % (ident, args)

        return "%s:%s" % (ident, self.pretty_type())


class JavaCodeInfo(object):
    """
    The 'Code' attribue of a method member of a java class

    reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.3
    """  # noqa

    def __init__(self, cpool):
        self.cpool = cpool
        self.attribs = JavaAttributes(cpool)
        self.max_stack = 0
        self.max_locals = 0
        self.code = None
        self.exceptions = tuple()

        # cache of disassembled code
        self._dis_code = None

        # cache of linenumbertable
        self._lnt = None


    def deref_const(self, index):
        """
        dereference a constant by index from the parent constant pool
        """

        return self.cpool.deref_const(index)


    def get_attribute(self, name):
        """
        get an attribute buffer by name
        """

        return self.attribs.get(name)


    def unpack(self, unpacker):
        """
        unpacks a code block from a buffer. Updates the internal structure
        of this instance
        """

        (a, b, c) = unpacker.unpack_struct(_HHI)

        self.max_stack = a
        self.max_locals = b
        self.code = unpacker.read(c)

        uobjs = unpacker.unpack_objects
        self.exceptions = tuple(uobjs(JavaExceptionInfo, self))

        self.attribs.unpack(unpacker)


    def get_linenumbertable(self):
        """
        a sequence of (code_offset, line_number) pairs.

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.12
        """  # noqa

        lnt = self._lnt
        if lnt is None:
            buff = self.get_attribute("LineNumberTable")
            if buff is None:
                lnt = tuple()
            else:
                with unpack(buff) as up:
                    lnt = tuple(up.unpack_struct_array(_HH))
            self._lnt = lnt
        return lnt


    def get_relativelinenumbertable(self):
        """
        a sequence of (code_offset, line_number) pairs. Similar to the
        get_linenumbertable method, but the line numbers start at 0
        (they are relative to the method, not to the class file)
        """

        lnt = self.get_linenumbertable()
        if lnt:
            lineoff = lnt[0][1]
            return tuple((o, l - lineoff) for (o, l) in lnt)
        else:
            return tuple()


    def get_localvariabletable(self):
        """
        a sequence of (code_offset, length, name_index, desc_index, index)
        tuples

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.13
        """  # noqa

        buff = self.get_attribute("LocalVariableTable")
        if buff is None:
            return tuple()

        with unpack(buff) as up:
            return tuple(up.unpack_struct_array(_HHHHH))


    def get_localvariabletypetable(self):
        """
        a sequence of (code_offset, length, name_index, signature_index,
        index) tuples

        reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.14
        """  # noqa

        buff = self.get_attribute("LocalVariableTypeTable")
        if buff is None:
            return tuple()

        with unpack(buff) as up:
            return tuple(up.unpack_struct_array(_HHHHH))


    def get_line_for_offset(self, code_offset):
        """
        returns the line number given a code offset
        """

        prev_line = 0

        for (offset, line) in self.get_linenumbertable():
            if offset < code_offset:
                prev_line = line
            elif offset == code_offset:
                return line
            else:
                return prev_line

        return prev_line


    def iter_code_by_lines(self):
        """
        ((abs_line, rel_line, [(offset, code, args), ...]),
         ...)
        """

        lnt = self.get_linenumbertable()
        if not lnt:
            yield (1, 1, self.disassemble())
            return

        lnt_offset = lnt[0][1]

        cur_line = None
        current = None

        for codelet in self.disassemble():
            abs_line = self.get_line_for_offset(codelet[0])

            if cur_line == abs_line:
                current.append(codelet)

            else:
                if cur_line is not None:
                    yield (cur_line, cur_line - lnt_offset, current)
                cur_line = abs_line
                current = [codelet]

        if current:
            yield (cur_line, cur_line - lnt_offset, current)


    def disassemble(self):
        """
        disassembles the underlying bytecode instructions and generates a
        sequence of (offset, code, args) tuples
        """

        dis = self._dis_code
        if dis is None:
            dis = tuple(disassemble(self.code))
            self._dis_code = dis

        return dis


class JavaExceptionInfo(object):
    """
    Information about an exception handler entry in an exception table
    """

    def __init__(self, code):
        self.code = code
        self.cpool = code.cpool

        self.start_pc = 0
        self.end_pc = 0
        self.handler_pc = 0
        self.catch_type_ref = 0


    def unpack(self, unpacker):
        """
        unpacks an exception handler entry in an exception table. Updates
        the internal structure of this instance
        """

        (a, b, c, d) = unpacker.unpack_struct(_HHHH)

        self.start_pc = a
        self.end_pc = b
        self.handler_pc = c
        self.catch_type_ref = d


    def get_catch_type(self):
        """
        dereferences the catch_type_ref to its class name, or None if the
        catch type is unspecified
        """

        if self.catch_type_ref:
            return self.cpool.deref_const(self.catch_type_ref)
        else:
            return None


    def pretty_catch_type(self):
        """
        pretty version of `get_catch_type()`

        If the catch type isn't specified, returns "any". Otherwise
        prefixes the pretty class name with the text "Class ". This is
        done to emulate the javap output for exceptions caught in a
        body of code.
        """

        ct = self.get_catch_type()
        if ct:
            return "Class " + _pretty_class(ct)
        else:
            return "any"


    def info(self):
        """
        tuple of the start_pc, end_pc, handler_pc and catch_type_ref
        """

        return (self.start_pc, self.end_pc,
                self.handler_pc, self.get_catch_type())


    def __eq__(self, other):
        return (isinstance(other, JavaExceptionInfo) and
                (self.info() == other.info()))


    def __ne__(self, other):
        return not self.__eq__(other)


    def __str__(self):
        return "(%s)" % ",".join(self.info())


class JavaInnerClassInfo(object):
    """
    Information about an inner class

    reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.6
    """  # noqa

    def __init__(self, cpool):
        self.cpool = cpool

        self.inner_info_ref = 0
        self.outer_info_ref = 0
        self.name_ref = 0
        self.access_flags = 0


    def unpack(self, unpacker):
        """
        unpack this instance with data from unpacker
        """

        (a, b, c, d) = unpacker.unpack_struct(_HHHH)

        self.inner_info_ref = a
        self.outer_info_ref = b
        self.name_ref = c
        self.access_flags = d


    def get_name(self):
        """
        the name of this inner-class
        """

        return self.cpool.deref_const(self.name_ref)


class JavaAnnotation(dict):
    """
    Java Annotations info

    reference: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html#jvms-4.7.16
    """  # noqa

    def __init__(self, cpool):
        dict.__init__(self)
        self.cpool = cpool
        self.type_ref = 0


    def unpack(self, unpacker):
        self.type_ref, count = unpacker.unpack_struct(_HH)

        for _i in range(0, count):
            key_ref, = unpacker.unpack_struct(_H)
            val = _unpack_annotation_val(unpacker, self.cpool)

            key = self.cpool.deref_const(key_ref)
            self[key] = val


    def pretty_type(self):
        return _pretty_type(self.cpool.deref_const(self.type_ref))


    def pretty_elements(self):
        result = list()

        for key, val in self.items():
            val = _pretty_annotation_val(val, self.cpool)
            result.append("%s=%s" % (key, val))

        return result


    def pretty_annotation(self):
        typename = self.pretty_type()
        elements = self.pretty_elements()
        return "%s(%s)" % (typename, ", ".join(elements))


    def __eq__(self, other):
        if not isinstance(other, JavaAnnotation):
            return False

        # if we have a different type name, not equal
        left = self.cpool.deref_const(self.type_ref)
        right = other.cpool.deref_const(other.type_ref)
        if left != right:
            return False

        # if we have differing sets of keys, not equal
        if self.keys() != other.keys():
            return False

        # for each of the key/val pairs, check equality
        for key, lval in self.items():
            rval = other[key]
            if not _annotation_val_eq(lval[0], lval[1], self.cpool,
                                      rval[0], rval[1], other.cpool):
                return False
        return True


    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '@' + self.pretty_annotation()


def _annotation_val_eq(left_tag, left_data, left_cpool,
                       right_tag, right_data, right_cpool):

    if left_tag != right_tag:
        return False

    lconst = left_cpool.deref_const
    rconst = right_cpool.deref_const

    if left_tag in 'BCDFIJSZs':
        return lconst(left_data) == rconst(right_data)

    elif left_tag == 'e':
        return ((lconst(left_data[0]) == rconst(right_data[0])) and
                (lconst(left_data[1]) == rconst(right_data[1])))

    elif left_tag == 'c':
        return lconst(left_data) == rconst(right_data)

    elif left_tag == '@':
        return left_data == right_data

    elif left_tag == '[':
        if len(left_data) != len(right_data):
            return False

        for index in range(0, len(left_data)):
            ld = left_data[index]
            rd = right_data[index]
            if not _annotation_val_eq(ld[0], ld[1], left_cpool,
                                      rd[0], rd[1], right_cpool):
                return False
        return True


def _unpack_annotation_val(unpacker, cpool):
    """
    tag, data tuple of an annotation
    """

    tag, = unpacker.unpack_struct(_B)
    tag = chr(tag)

    if tag in 'BCDFIJSZs':
        data, = unpacker.unpack_struct(_H)

    elif tag == 'e':
        data = unpacker.unpack_struct(_HH)

    elif tag == 'c':
        data, = unpacker.unpack_struct(_H)

    elif tag == '@':
        data = JavaAnnotation(cpool)
        data.unpack(unpacker)

    elif tag == '[':
        data = list()
        count, = unpacker.unpack_struct(_H)
        for _i in range(0, count):
            data.append(_unpack_annotation_val(unpacker, cpool))

    else:
        raise Unimplemented("Unknown tag {}".format(tag))

    return tag, data


def _pretty_annotation_val(val, cpool):
    """
    a pretty display of a tag and data pair annotation value
    """

    tag, data = val

    if tag in 'BCDFIJSZs':
        data = "%s#%i" % (tag, data)

    elif tag == 'e':
        data = "e#%i.#%i" % data

    elif tag == 'c':
        data = "c#%i" % data

    elif tag == '@':
        data = "@" + data.pretty_annotation()

    elif tag == '[':
        combine = list()
        for val in data:
            combine.append(_pretty_annotation_val(val, cpool))
        data = "[%s]" % ", ".join(combine)

    return data


# -----
# Utility functions for turning major/minor versions into JVM releases
# Each entry is a tuple of minimum version and maxiumum version,
# inclusive, and the string of the platform version.


_platforms = (
    ((45, 0), (45, 3), "1.0.2"),
    ((45, 4), (45, 65535), "1.1"),
    ((46, 0), (46, 65535), "1.2"),
    ((47, 0), (47, 65535), "1.3"),
    ((48, 0), (48, 65535), "1.4"),
    ((49, 0), (49, 65535), "1.5"),
    ((50, 0), (50, 65535), "1.6"),
    ((51, 0), (51, 65535), "1.7"),
    ((52, 0), (52, 65535), "1.8"), )


def platform_from_version(major, minor):
    """
    returns the minimum platform version that can load the given class
    version indicated by major.minor or None if no known platforms
    match the given version
    """

    v = (major, minor)
    for low, high, name in _platforms:
        if low <= v <= high:
            return name
    return None


# -----
# Utility functions for the constants pool


def _unpack_const_item(unpacker):
    """
    unpack a constant pool item, which will consist of a type byte
    (see the CONST_ values in this module) and a value of the
    appropriate type
    """

    (typecode,) = unpacker.unpack_struct(_B)

    if typecode == CONST_Utf8:
        (slen,) = unpacker.unpack_struct(_H)
        val = unpacker.read(slen)
        try:
            val = val.decode("utf8")
        except UnicodeDecodeError:
            # easiest hack to handle java's modified utf-8 encoding
            # also we want at least some data, thus we ignore
            # unknown characters
            val = val.replace(b"\xC0\x80", b"\x00") \
                     .decode("utf8", errors="ignore")

    elif typecode == CONST_Integer:
        (val,) = unpacker.unpack(">i")

    elif typecode == CONST_Float:
        (val,) = unpacker.unpack(">f")

    elif typecode == CONST_Long:
        (val,) = unpacker.unpack(">q")

    elif typecode == CONST_Double:
        (val,) = unpacker.unpack(">d")

    elif typecode in (CONST_Class, CONST_String, CONST_MethodType,
                      CONST_Module, CONST_Package):
        (val,) = unpacker.unpack_struct(_H)

    elif typecode in (CONST_Fieldref, CONST_Methodref,
                      CONST_InterfaceMethodref, CONST_NameAndType,
                      CONST_ModuleId, CONST_InvokeDynamic, CONST_Dynamic):
        val = unpacker.unpack_struct(_HH)

    elif typecode == CONST_MethodHandle:
        val = unpacker.unpack_struct(_BH)

    elif typecode == CONST_MethodType:
        val = unpacker.unpack_struct(_H)

    else:
        raise UnknownConstantPoolTagException(
            "unknown constant type %r" % typecode)

    return typecode, val


def _pretty_const_type_val(typecode, val):
    """
    given a typecode and a value, returns the appropriate pretty
    version of that value (not the dereferenced data)
    """

    if typecode == CONST_Utf8:
        typestr = "Utf8"  # formerly Asciz, which was considered Java bug
        if not isinstance(val, str):  # Py2, val is 'unicode'
            val = repr(val)[2:-1]  # trim off the surrounding u"" (HACK)
        else:
            val = repr(val)[1:-1]  # trim off the surrounding "" (HACK)
    elif typecode == CONST_Integer:
        typestr = "int"
    elif typecode == CONST_Float:
        typestr = "float"
        val = "%ff" % val
    elif typecode == CONST_Long:
        typestr = "long"
        val = "%il" % val
    elif typecode == CONST_Double:
        typestr = "double"
        val = "%fd" % val
    elif typecode == CONST_Class:
        typestr = "class"
        val = "#%i" % val
    elif typecode == CONST_String:
        typestr = "String"
        val = "#%i" % val
    elif typecode == CONST_Fieldref:
        typestr = "Field"
        val = "#%i.#%i" % val
    elif typecode == CONST_Methodref:
        typestr = "Method"
        val = "#%i.#%i" % val
    elif typecode == CONST_InterfaceMethodref:
        typestr = "InterfaceMethod"
        val = "#%i.#%i" % val
    elif typecode == CONST_NameAndType:
        typestr = "NameAndType"
        val = "#%i:#%i" % val
    elif typecode == CONST_ModuleId:
        typestr = "ModuleId"
        val = "#%i@#%i" % val
    elif typecode == CONST_MethodHandle:
        typestr = "MethodHandle"
        val = repr(val)
    elif typecode == CONST_MethodType:
        typestr = "MethodType"
        val = repr(val)
    elif typecode == CONST_InvokeDynamic:
        typestr = "InvokeDynamic"
        val = repr(val)
    elif typecode == CONST_Module:
        typestr = "Module"
    elif typecode == CONST_Package:
        typestr = "Package"
    else:
        raise UnknownConstantPoolTagException(
            "unknown constant type %r" % typecode)

    return typestr, val


# -----
# Utility functions for dealing with exploding internal type
# signatures into sequences, and converting type signatures into
# "pretty" strings


def pretty_generic(signature):
    """
    Pretty version of the given generics signature
    """

    # TODO: this format is annoying in so many ways.

    return signature


def _next_argsig(s):
    """
    given a string, find the next complete argument signature and
    return it and a new string advanced past that point
    """

    c = s[0]

    if c in "BCDFIJSVZ":
        result = (c, s[1:])

    elif c == "[":
        d, s = _next_argsig(s[1:])
        result = (c + d, s)

    elif c == "L":
        i = s.find(';') + 1
        result = (s[:i], s[i:])

    elif c == "(":
        i = s.find(')') + 1
        result = (s[:i], s[i:])

    # Some files may be corrupted and contain bad argument signature (c = '.').
    # We do not want to fail on them rather skip this argument signature,
    # continue and finish successfully.
    # Example:
    # s = '[Lcom.sun.glass.ui.EventLoop$State;.clone():java.lang.Object'
    elif c == ".":
        result = (s[1:], "")

    else:
        raise Unimplemented("_next_argsig is %r in %r" % (c, s))

    return result


def _typeseq_iter(s):
    """
    iterate through all of the type signatures in a sequence
    """

    original = s
    try:
        s = str(s)
        while s:
            t, s = _next_argsig(s)

            # If `original == s` then it is possible that argument signature
            # is corrupted which may lead to an infinite loop.
            # In this case, raise an exception.
            # Example:
            # original = "LFR1Xdn"
            # t, s = _next_argsig("LFR1Xdn"); original == s
            if original == s and t == "":
                raise Unimplemented(
                    "Infinite loop detected: %r, %r = _next_argsig(%r)" %
                    (t, s, original)
                )

            yield t

    except Unimplemented:
        raise Unimplemented("Unknown type signature in %r" % original)


def _typeseq(type_s):
    """
    tuple version of _typeseq_iter
    """

    return tuple(_typeseq_iter(type_s))


def _pretty_typeseq(type_s):
    """
    iterator of pretty versions of _typeseq_iter
    """

    return (_pretty_type(t) for t in _typeseq(type_s))


def _pretty_type(s, offset=0):
    # pylint: disable=R0911, R0912
    # too many returns, too many branches. Not converting this to a
    # dict lookup. Waiving instead.

    """
    returns the pretty version of a type code
    """

    tc = s[offset]

    if tc == "V":
        return "void"

    elif tc == "Z":
        return "boolean"

    elif tc == "C":
        return "char"

    elif tc == "B":
        return "byte"

    elif tc == "S":
        return "short"

    elif tc == "I":
        return "int"

    elif tc == "J":
        return "long"

    elif tc == "D":
        return "double"

    elif tc == "F":
        return "float"

    elif tc == "L":
        return _pretty_class(s[offset + 1:-1])

    elif tc == "[":
        return "%s[]" % _pretty_type(s, offset + 1)

    elif tc == "(":
        return "(%s)" % ",".join(_pretty_typeseq(s[offset + 1:-1]))

    elif tc == "T":
        return "generic " + s[offset + 1:]

    else:
        raise Unimplemented("unknown type, %r" % tc)


def _pretty_class(s):
    """
    convert the internal class name representation into what users
    expect to see. Currently that just means swapping '/' for '.'
    """

    # well that's easy.
    return s.replace("/", ".")


# -----
# Functions for dealing with buffers and files


def is_class(data):
    """
    checks that the data (which is a string, buffer, or a stream
    supporting the read method) has the magic numbers indicating it is
    a Java class file. Returns False if the magic numbers do not
    match, or for any errors.
    """

    try:
        with unpack(data) as up:
            magic = up.unpack_struct(_BBBB)

        return magic == JAVA_CLASS_MAGIC

    except UnpackException:
        return False


def is_class_file(filename):
    """
    checks whether the given file is a Java class file, by opening it
    and checking for the magic header
    """

    with open(filename, "rb") as fd:
        c = fd.read(len(JAVA_CLASS_MAGIC))
        if isinstance(c, str):      # Python 2
            c = map(ord, c)
        return tuple(c) == JAVA_CLASS_MAGIC


def unpack_class(data, magic=None):
    """
    unpacks a Java class from data, which can be a string, a buffer,
    or a stream supporting the read method. Returns a populated
    JavaClassInfo instance.

    If data is a stream which has already been confirmed to be a java
    class, it may have had the first four bytes read from it already.
    In this case, pass those magic bytes as a str or tuple and the
    unpacker will not attempt to read them again.

    Raises a ClassUnpackException or an UnpackException if the class
    data is malformed. Raises Unimplemented if a feature is discovered
    which isn't understood by javatools yet.
    """

    with unpack(data) as up:
        magic = magic or up.unpack_struct(_BBBB)
        if magic != JAVA_CLASS_MAGIC:
            raise ClassUnpackException("Not a Java class file")

        o = JavaClassInfo()
        o.unpack(up, magic=magic)

    return o


def unpack_classfile(filename):
    """
    returns a newly allocated JavaClassInfo object populated with the
    data unpacked from the specified file. Raises an UnpackException
    if the class data is malformed
    """

    with open(filename, "rb", _BUFFERING) as fd:
        return unpack_class(fd.read())


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Cryptography-related functions for handling JAR signature block files.

:author: Konstantin Shemyak  <konstantin@shemyak.com>
:license: LGPL v.3
"""


from M2Crypto import SMIME, X509, BIO, RSA, DSA, EC, m2


class CannotFindKeyTypeError(Exception):
    """
    Failed to determine the type of the private key.
    """

    pass


class SignatureBlockVerificationError(Exception):
    """
    The Signature Block File verification failed.
    """

    pass


def private_key_type(key_file):
    """
    Determines type of the private key: RSA, DSA, EC.

    :param key_file: file path
    :type key_file: str
    :return: one of "RSA", "DSA" or "EC"
    :except CannotFindKeyTypeError
    """

    keytypes = (("RSA", RSA), ("DSA", DSA), ("EC", EC))

    for key, ktype in keytypes:
        try:
            ktype.load_key(key_file)
        except (RSA.RSAError, DSA.DSAError, ValueError):
            continue
        else:
            return key
    else:
        raise CannotFindKeyTypeError()


def create_signature_block(openssl_digest, certificate, private_key,
                           extra_certs, data):

    """
    Produces a signature block for the data.

    Reference
    ---------
    http://docs.oracle.com/javase/7/docs/technotes/guides/jar/jar.html#Digital_Signatures

    Note: Oracle does not specify the content of the "signature
    file block", friendly saying that "These are binary files
    not intended to be interpreted by humans".

    :param openssl_digest: alrogithm known to OpenSSL used to digest the data
    :type openssl_digest: str
    :param certificate: filename of the certificate file (PEM format)
    :type certificate: str
    :param private_key:filename of private key used to sign (PEM format)
    :type private_key: str
    :param extra_certs: additional certificates to embed into the signature (PEM format)
    :type extra_certs: array of filenames
    :param data: the content to be signed
    :type data: bytes
    :returns: content of the signature block file as produced by jarsigner
    :rtype: bytes
    """  # noqa

    smime = SMIME.SMIME()
    with BIO.openfile(private_key) as k, BIO.openfile(certificate) as c:
        smime.load_key_bio(k, c)

    if extra_certs is not None:
        # Could we use just X509.new_stack_from_der() instead?
        stack = X509.X509_Stack()
        for cert in extra_certs:
            stack.push(X509.load_cert(cert))
        smime.set_x509_stack(stack)

    pkcs7 = smime.sign(BIO.MemoryBuffer(data),
                       algo=openssl_digest,
                       flags=(SMIME.PKCS7_BINARY |
                              SMIME.PKCS7_DETACHED |
                              SMIME.PKCS7_NOATTR))
    tmp = BIO.MemoryBuffer()
    pkcs7.write_der(tmp)
    return tmp.read()


def ignore_missing_email_protection_eku_cb(ok, ctx):
    """
    For verifying PKCS7 signature, m2Crypto uses OpenSSL's PKCS7_verify().
    The latter requires that ExtendedKeyUsage extension, if present,
    contains 'emailProtection' OID. (Is it because S/MIME is/was the
    primary use case for PKCS7?)
    We do not want to fail the verification in this case. At present,
    M2Crypto lacks possibility of removing or modifying an existing
    extension. Let's assign a custom verification callback.
    """
    # The error we want to ignore is indicated by X509_V_ERR_INVALID_PURPOSE.
    err = ctx.get_error()
    if err != m2.X509_V_ERR_INVALID_PURPOSE:
        return ok

    # PKCS7_verify() has this requriement only for the signing certificate.
    # Do not modify the behavior for certificates upper in the chain.
    if ctx.get_error_depth() > 0:
        return ok

    # There is another cause of ERR_INVALID_PURPOSE: incompatible keyUsage.
    # Do not modify the default behavior in this case.
    cert = ctx.get_current_cert()
    try:
        key_usage = cert.get_ext('keyUsage').get_value()
        if 'digitalSignature' not in key_usage \
                and 'nonRepudiation' not in key_usage:
            return ok
    except LookupError:
        pass

    # Here, keyUsage is either absent, or contains the needed bit(s).
    # So ERR_INVALID_PURPOSE is caused by EKU not containing 'emailProtection'.
    # Ignore this error.
    return 1


def verify_signature_block(certificate_file, content, signature):
    """
    Verifies the 'signature' over the 'content', trusting the
    'certificate'.

    :param certificate_file: the trusted certificate (PEM format)
    :type certificate_file: str
    :param content: The signature should match this content
    :type content: str
    :param signature: data (DER format) subject to check
    :type signature: str
    :return None if the signature validates.
    :exception SignatureBlockVerificationError
    """

    sig_bio = BIO.MemoryBuffer(signature)
    pkcs7 = SMIME.PKCS7(m2.pkcs7_read_bio_der(sig_bio._ptr()), 1)
    signers_cert_stack = pkcs7.get0_signers(X509.X509_Stack())
    trusted_cert_store = X509.X509_Store()
    trusted_cert_store.set_verify_cb(ignore_missing_email_protection_eku_cb)
    trusted_cert_store.load_info(certificate_file)
    smime = SMIME.SMIME()
    smime.set_x509_stack(signers_cert_stack)
    smime.set_x509_store(trusted_cert_store)
    data_bio = BIO.MemoryBuffer(content)

    try:
        smime.verify(pkcs7, data_bio)
    except SMIME.PKCS7_Error as message:
        raise SignatureBlockVerificationError(message)
    else:
        return None


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Module and utility for creating, modifying, signing, or verifying
Java archives

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""


from __future__ import print_function


import argparse
import os
import sys

from shutil import copyfile
from tempfile import NamedTemporaryFile
from zipfile import ZipFile, ZIP_DEFLATED

from .manifest import file_matches_sigfile, Manifest, SignatureManifest
from .crypto import private_key_type, CannotFindKeyTypeError
from .crypto import verify_signature_block, SignatureBlockVerificationError


__all__ = (
    "cli_create_jar", "cli_sign_jar",
    "cli_verify_jar_signature", "main", )


class VerificationError(Exception):
    pass


class SignatureBlockFileVerificationError(VerificationError):
    pass


class ManifestChecksumError(VerificationError):
    pass


class JarChecksumError(VerificationError):
    pass


class JarSignatureMissingError(VerificationError):
    pass


class MissingManifestError(Exception):
    pass


def verify(certificate, jar_file, sf_name=None):
    """
    Verifies signature of a JAR file.

    Limitations:
    - diagnostic is less verbose than of jarsigner
    :return None if verification succeeds.
    :exception SignatureBlockFileVerificationError, ManifestChecksumError,
        JarChecksumError, JarSignatureMissingError

    Reference:
    http://docs.oracle.com/javase/7/docs/technotes/guides/jar/jar.html#Signature_Validation
    Note that the validation is done in three steps. Failure at any step is a
    failure of the whole validation.
    """  # noqua

    # Step 0: get the "key alias", used also for naming of sig-related files.
    zip_file = ZipFile(jar_file)
    sf_files = [f for f in zip_file.namelist() if file_matches_sigfile(f)]

    if len(sf_files) == 0:
        raise JarSignatureMissingError("No .SF file in %s" % jar_file)

    elif len(sf_files) > 1:
        if sf_name is None:
            msg = "Multiple .SF files in %s, but SF_NAME.SF not specified" \
                % jar_file
            raise VerificationError(msg)

        elif ('META-INF/' + sf_name) in sf_files:
            sf_filename = 'META-INF/' + sf_name

        else:
            msg = "No .SF file in %s named META-INF/%s (found %d .SF files)" \
                % (jar_file, sf_name, len(sf_files))
            raise VerificationError(msg)

    elif len(sf_files) == 1:
        if sf_name is None:
            sf_filename = sf_files[0]

        elif sf_files[0] == 'META-INF/' + sf_name:
            sf_filename = sf_files[0]

        else:
            msg = "No .SF file in %s named META-INF/%s" % (jar_file, sf_name)
            raise VerificationError(msg)

    key_alias = sf_filename[9:-3]  # "META-INF/%s.SF"
    sf_data = zip_file.read(sf_filename)

    # Step 1: check the crypto part.
    file_list = zip_file.namelist()
    sig_block_filename = None

    # JAR specification mentions only RSA and DSA; jarsigner also has EC
    # TODO: what about "SIG-*"?
    signature_extensions = ("RSA", "DSA", "EC")
    for extension in signature_extensions:
        candidate_filename = "META-INF/%s.%s" % (key_alias, extension)
        if candidate_filename in file_list:
            sig_block_filename = candidate_filename
            break

    if sig_block_filename is None:
        msg = "None of %s found in JAR" % \
              ", ".join(key_alias + "." + x for x in signature_extensions)
        raise JarSignatureMissingError(msg)

    sig_block_data = zip_file.read(sig_block_filename)
    try:
        verify_signature_block(certificate, sf_data, sig_block_data)

    except SignatureBlockVerificationError as message:
        message = "Signature block verification failed: %s" % message
        raise SignatureBlockFileVerificationError(message)

    # KEYALIAS.SF is correctly signed.
    # Step 2: Check that it contains correct checksum of the manifest.
    signature_manifest = SignatureManifest()
    signature_manifest.parse(sf_data)

    jar_manifest = Manifest()
    jar_manifest.load_from_jar(jar_file)

    errors = signature_manifest.verify_manifest(jar_manifest)
    if len(errors) > 0:
        msg = "%s: in .SF file, section checksum(s) failed for: %s" \
              % (jar_file, ",".join(errors))
        raise ManifestChecksumError(msg)

    # Checksums of MANIFEST.MF itself are correct.

    # Step 3: Check that it contains valid checksums for each file
    # from the JAR.  NOTE: the check is done for JAR entries. If some
    # JAR entries are deleted after signing, the verification still
    # succeeds.  This seems to not follow the reference specification,
    # but that's what jarsigner does.
    errors = jar_manifest.verify_jar_checksums(jar_file)
    if len(errors) > 0:
        msg = "Checksum(s) for jar entries of jar file %s failed for: %s" \
              % (jar_file, ",".join(errors))
        raise JarChecksumError(msg)

    return None


def sign(jar_file, cert_file, key_file, key_alias,
         extra_certs=None, digest="SHA-256", output=None):
    """
    Signs the jar (almost) identically to jarsigner.
    :exception ManifestNotFoundError, CannotFindKeyTypeError
    :return None
    """

    mf = Manifest()
    mf.load_from_jar(jar_file)
    mf.add_jar_entries(jar_file, digest)

    # create a signature manifest, and make it match the line separator
    # style of the manifest it'll be digesting.
    sf = SignatureManifest(linesep=mf.linesep)

    sf_digest_algorithm = "SHA-256" if digest is None else digest
    sf.digest_manifest(mf, sf_digest_algorithm)

    sig_digest_algorithm = digest  # No point to make it different
    sig_block_extension = private_key_type(key_file)

    sigdata = sf.get_signature(cert_file, key_file,
                               extra_certs, sig_digest_algorithm)

    # We might just add new entries to the original JAR, but jarsigner puts
    # all META-INF/ to the beginning of the archive. Let's do the same.
    with NamedTemporaryFile() as new_jar_file:
        new_jar = ZipFile(new_jar_file, "w", ZIP_DEFLATED)
        new_jar.writestr("META-INF/MANIFEST.MF", mf.get_data())
        new_jar.writestr("META-INF/%s.SF" % key_alias, sf.get_data())
        new_jar.writestr("META-INF/%s.%s" % (key_alias, sig_block_extension),
                         sigdata)
        jar = ZipFile(jar_file, "a")
        for entry in jar.namelist():
            # TODO: In Py2, namelist() can be of type unicode
            if not entry.upper() == "META-INF/MANIFEST.MF":
                new_jar.writestr(entry, jar.read(entry))

        new_jar.close()
        new_jar_file.flush()
        copyfile(new_jar_file.name, jar_file if output is None else output)


def create_jar(jar_file, entries):
    """
    Create JAR from given entries.
    :param jar_file: filename of the created JAR
    :type jar_file: str
    :param entries: files to put into the JAR
    :type entries: list[str]
    :return: None
    """

    # 'jar' adds separate entries for directories, also for empty ones.
    with ZipFile(jar_file, "w") as jar:
        jar.writestr("META-INF/", "")
        jar.writestr("META-INF/MANIFEST.MF", Manifest().get_data())
        for entry in entries:
            jar.write(entry)
            if os.path.isdir(entry):
                for root, dirs, files in os.walk(entry):
                    for filename in dirs + files:
                        jar.write(os.path.join(root, filename))


def cli_create_jar(argument_list):
    """
    A subset of "jar" command. Creating new JARs only.
    """

    usage_message = "usage: jarutil c [OPTIONS] file.jar files..."
    parser = argparse.ArgumentParser(usage=usage_message)

    parser.add_argument("jar_file", type=str,
                        help="The file to create")
    parser.add_argument("-m", "--main-class", type=str,
                        help="Specify application entry point")

    args, entries = parser.parse_known_args(argument_list)
    create_jar(args.jar_file, entries)
    return 0


def cli_sign_jar(argument_list=None):
    """
    Command-line wrapper around sign()
    """

    usage_message = "jarutil s [OPTIONS] jar_file cert_file " \
                    "key_file key_alias"

    parser = argparse.ArgumentParser(usage=usage_message)

    parser.add_argument("jar_file", type=str,
                        help="JAR file to sign")
    parser.add_argument("cert_file", type=str,
                        help="Certificate file (PEM format)")
    parser.add_argument("key_file", type=str,
                        help="Private key file (PEM format)")
    parser.add_argument("key_alias", type=str,
                        help="JAR \"key alias\" to use")

    parser.add_argument("-d", "--digest", action="store", default="SHA-256",
                        # TODO: factor to a constant
                        help="Digest algorithm used for signing")

    parser.add_argument("-c", "--chain", action="append",
                        dest="extra_certs", default=[],
                        help="Additional certificates to embed into the"
                        " signature (PEM format). More than one \"-c\" option"
                        " can be provided.")

    parser.add_argument("-o", "--output", action="store", default=None,
                        help="Filename to put signed jar. If not provided, the"
                        " signature is added to the original jar file.")

    args = parser.parse_args(argument_list)

    try:
        sign(args.jar_file, args.cert_file, args.key_file, args.key_alias,
             args.extra_certs, args.digest, args.output)

    except CannotFindKeyTypeError:
        print("Cannot determine private key type in %s" % args.key_file)
        return 1

    except MissingManifestError:
        print("Manifest missing in jar file %s" % args.jar_file)
        return 2

    return 0


def cli_verify_jar_signature(argument_list):
    """
    Command-line wrapper around verify()
    TODO: use trusted keystore;
    """

    usage_message = "jarutil v file.jar trusted_certificate.pem [SF_NAME.SF]"
    if len(argument_list) < 2 or len(argument_list) > 3:
        print(usage_message)
        return 1

    jar_file, certificate, sf_name = (argument_list + [None])[:3]
    try:
        verify(certificate, jar_file, sf_name)

    except VerificationError as error_message:
        print(error_message)
        return 1

    else:
        print("Jar verified.")
        return 0


def usage():
    # TODO, use argparse/optionparser to do this

    print("Usage: jarutil command [options] [argument]...")
    print("Commands:")
    print("   c: create JAR from paths")
    print("   s: sign JAR")
    print("   v: verify JAR signature")
    print("Give option \"-h\" for help on a particular command.")

    return 1


def main(args=sys.argv):
    if len(args) < 2:
        return usage()

    command = args[1]
    rest = args[2:]

    if "create".startswith(command):
        return cli_create_jar(rest)
    elif "sign".startswith(command):
        return cli_sign_jar(rest)
    elif "view".startswith(command):
        return cli_verify_jar_signature(rest)
    else:
        return usage()


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.



"""
Utility module for unpacking shapes of binary data from a buffer
or stream.

:author: Christopher O'Brien <obriencj@gmail.com>
:license: LGPL v.3
"""


# TODO: maybe one day in the future we'll add a Packer to this. If we
# ever get to the point where we want to recompile a class from a
# JavaClassInfo instance.

# profiling showed a significant amount of time spent in this module,
# so there will be efforts here to increase performance


from abc import ABCMeta, abstractmethod
from six import add_metaclass
from six.moves import range
from struct import Struct


__all__ = (
    "compile_struct", "unpack",
    "Unpacker", "UnpackException",
    "StreamUnpacker", "BufferUnpacker",
)

try:
    buffer
except NameError:
    buffer = memoryview

# pylint: disable=C0103
_struct_cache = dict()


def compile_struct(fmt, cache=None):
    """
    returns a struct.Struct instance compiled from fmt. If fmt has
    already been compiled, it will return the previously compiled
    Struct instance from the cache.
    """

    if cache is None:
        cache = _struct_cache

    sfmt = cache.get(fmt, None)
    if not sfmt:
        sfmt = Struct(fmt)
        cache[fmt] = sfmt
    return sfmt


@add_metaclass(ABCMeta)
class Unpacker(object):
    """
    Abstract base class for `StreamUnpacker` and `BufferUnpacker`. Use
    the `unpack` function to obtain the correct unpacker instance for
    your data.
    """


    def __enter__(self):
        return self


    def __exit__(self, exc_type, _exc_val, _exc_tb):
        self.close()
        return exc_type is None


    @abstractmethod
    def unpack(self, format_str):  # pragma: no cover
        """
        unpacks the given format_str from the underlying data and returns
        the results. Will raise an UnpackException if there is not
        enough data to satisfy the specified format
        """

        pass


    @abstractmethod
    def unpack_struct(self, struct):  # pragma: no cover
        """
        unpacks the given struct from the underlying data and returns the
        results. Will raise an UnpackException if there is not enough
        data to satisfy the format of the structure
        """

        pass


    @abstractmethod
    def read(self, count):  # pragma: no cover
        """
        read count bytes from the unpacker and return it. Raises an
        UnpackException if there is not enough data in the underlying
        stream.
        """

        pass


    @abstractmethod
    def close(self):  # pragma: no cover
        """
        close this unpacker and release the underlying data
        """

        pass


    def unpack_array(self, fmt):
        """
        reads a count from the unpacker, and unpacks fmt count
        times. Yields a sequence of the unpacked data tuples
        """

        (count,) = self.unpack_struct(_H)
        sfmt = compile_struct(fmt)
        for _i in range(count):
            yield self.unpack_struct(sfmt)


    def unpack_struct_array(self, struct):
        """
        reads a count from the unpacker, and unpacks the precompiled
        struct count times. Yields a sequence of the unpacked data
        tuples
        """

        (count,) = self.unpack_struct(_H)
        for _i in range(count):
            yield self.unpack_struct(struct)


    def unpack_objects(self, atype, *params, **kwds):
        """
        reads a count from the unpacker, and instanciates that many calls
        to atype, with the given params and kwds passed along. Each
        instance then has its unpack method called with this unpacker
        instance passed along. Yields a squence of the unpacked
        instances
        """

        (count,) = self.unpack_struct(_H)
        for _i in range(count):
            obj = atype(*params, **kwds)
            obj.unpack(self)
            yield obj


class BufferUnpacker(Unpacker):
    """
    Unpacker wrapping a str or buffer.
    """

    def __init__(self, data, offset=0):
        super(BufferUnpacker, self).__init__()
        self.data = data
        self.offset = offset


    def unpack(self, fmt):
        """
        unpacks the given fmt from the underlying buffer and returns the
        results. Will raise an UnpackException if there is not enough
        data to satisfy the fmt
        """

        sfmt = compile_struct(fmt)
        size = sfmt.size

        offset = self.offset
        if self.data:
            avail = len(self.data) - offset
        else:
            avail = 0

        if avail < size:
            raise UnpackException(fmt, size, avail)

        self.offset = offset + size
        return sfmt.unpack_from(self.data, offset)


    def unpack_struct(self, struct):
        """
        unpacks the given struct from the underlying buffer and returns
        the results. Will raise an UnpackException if there is not
        enough data to satisfy the format of the structure
        """

        size = struct.size

        offset = self.offset
        if self.data:
            avail = len(self.data) - offset
        else:
            avail = 0

        if avail < size:
            raise UnpackException(struct.format, size, avail)

        self.offset = offset + size
        return struct.unpack_from(self.data, offset)


    def read(self, count):
        """
        read count bytes from the underlying buffer and return them as a
        str. Raises an UnpackException if there is not enough data in
        the underlying buffer.
        """

        offset = self.offset
        if self.data:
            avail = len(self.data) - offset
        else:
            avail = 0

        if avail < count:
            raise UnpackException(None, count, avail)

        self.offset = offset + count
        return self.data[offset:self.offset]


    def close(self):
        """
        release the underlying buffer
        """

        self.data = None
        self.offset = 0


class StreamUnpacker(Unpacker):
    """
    Wraps a stream (or creates a stream for a string or buffer) and
    advances along it while unpacking structures from it.

    This class adheres to the context management protocol, so may be
    used in conjunction with the 'with' keyword
    """

    def __init__(self, data):
        super(StreamUnpacker, self).__init__()
        self.data = data


    def unpack(self, fmt):
        """
        unpacks the given fmt from the underlying stream and returns the
        results. Will raise an UnpackException if there is not enough
        data to satisfy the fmt
        """

        sfmt = compile_struct(fmt)
        size = sfmt.size

        if not self.data:
            raise UnpackException(fmt, size, 0)

        buff = self.data.read(size)
        if len(buff) < size:
            raise UnpackException(fmt, size, len(buff))

        return sfmt.unpack(buff)


    def unpack_struct(self, struct):
        """
        unpacks the given struct from the underlying stream and returns
        the results. Will raise an UnpackException if there is not
        enough data to satisfy the format of the structure
        """

        size = struct.size

        if not self.data:
            raise UnpackException(struct.format, size, 0)

        buff = self.data.read(size)
        if len(buff) < size:
            raise UnpackException(struct.format, size, len(buff))

        return struct.unpack(buff)


    def read(self, count):
        """
        read count bytes from the unpacker and return it. Raises an
        UnpackException if there is not enough data in the underlying
        stream.
        """

        if not self.data:
            raise UnpackException(None, count, 0)

        buff = self.data.read(count)
        if len(buff) < count:
            raise UnpackException(None, count, len(buff))

        return buff


    def close(self):
        """
        close this unpacker, and the underlying stream if it supports such
        """

        data = self.data
        self.data = None

        if hasattr(data, "close"):
            data.close()


def unpack(data):
    """
    returns either a BufferUnpacker or StreamUnpacker instance,
    depending upon the type of data. The unpacker supports the managed
    context interface, so may be used eg: `with unpack(my_data) as
    unpacker:`
    """

    if isinstance(data, (bytes, buffer)):
        return BufferUnpacker(data)

    elif hasattr(data, "read"):
        return StreamUnpacker(data)

    else:
        raise TypeError("unpack requires bytes, buffer, or instance"
                        " supporting the read method")


class UnpackException(Exception):
    """
    raised when there is not enough data to unpack the expected
    structures
    """

    template = "format %r requires %i bytes, only %i present"


    def __init__(self, fmt, wanted, present):
        msg = self.template % (fmt, wanted, present)
        super(UnpackException, self).__init__(msg)

        self.format = fmt
        self.bytes_wanted = wanted
        self.bytes_present = present


# We use this one a lot, so let's not bother calling compile_struct
# over and over to get it.
_H = compile_struct(">H")


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Utility for comparing two distributions.

Distributions are directories. Any JAR files (eg: jar, sar, ear, war)
will be deeply checked for their class members. And Class files will
be checked for deep differences.

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""


import sys

from argparse import ArgumentParser
from multiprocessing import cpu_count
from os.path import join
from six.moves import range, zip_longest

from . import unpack_classfile
from .change import GenericChange, SuperChange, Addition, Removal
from .change import squash, yield_sorted_by_type
from .classdiff import JavaClassChange, JavaClassReport
from .classdiff import add_classdiff_optgroup, add_general_optgroup
from .dirutils import compare, fnmatches
from .dirutils import LEFT, RIGHT, SAME, DIFF
from .manifest import Manifest, ManifestChange
from .jardiff import JarChange, JarReport, add_jardiff_optgroup
from .jarinfo import JAR_PATTERNS


__all__ = (
    "TEXT_PATTERNS",
    "DistChange",
    "DistContentChange", "DistContentAdded", "DistContentRemoved",
    "DistTextChange", "DistManifestChange",
    "DistClassChange", "DistClassAdded", "DistClassRemoved",
    "DistJarChange", "DistJarAdded", "DistJarRemoved",
    "DistReport", "DistClassReport", "DistJarReport",
    "cli", "main",
    "cli_dist_diff", "default_distdiff_options", )


# glob patterns used to trigger a DistTextChange
TEXT_PATTERNS = (
    "*.bat",
    "*.cert",
    "*.cfg",
    "*.conf",
    "*.dtd",
    "*.html",
    "*.ini",
    "*.properties",
    "*.sh",
    "*.text",
    "*.txt",
    "*.xml", )


class DistContentChange(SuperChange):

    label = "Distributed Content"


    def __init__(self, ldir, rdir, entry, change=True):
        super(DistContentChange, self).__init__(ldir, rdir)
        self.entry = entry
        self.changed = change


    def left_fn(self):
        return join(self.ldata, self.entry)


    def right_fn(self):
        return join(self.rdata, self.entry)


    def open_left(self, mode="rb"):
        return open(self.left_fn(), mode)


    def open_right(self, mode="rb"):
        return open(self.right_fn(), mode)


    def collect_impl(self):

        """ Content changes refer to more concrete children, but by
        default are empty """

        return tuple()


    def get_description(self):
        c = "has changed" if self.is_change() else "is unchanged"
        return "%s %s: %s" % (self.label, c, self.entry)


    def is_ignored(self, options):
        return (fnmatches(self.entry, *options.ignore_filenames) or
                SuperChange.is_ignored(self, options))


class DistContentAdded(DistContentChange, Addition):

    label = "Distributed Content Added"


    def get_description(self):
        return "%s: %s" % (self.label, self.entry)


class DistContentRemoved(DistContentChange, Removal):

    label = "Distributed Content Removed"


    def get_description(self):
        return "%s: %s" % (self.label, self.entry)


class DistTextChange(DistContentChange):

    label = "Distributed Text"


    def __init__(self, l, r, entry, change=True):
        super(DistTextChange, self).__init__(l, r, entry, change)
        self.lineending = False


    def check(self):
        # if the file matches what we would consider a text file,
        # check if the only difference is in the trailing whitespace,
        # and if so, set lineending to true so we can optionally
        # ignore the change later.
        with self.open_left(mode="rt") as lfd:
            with self.open_right(mode="rt") as rfd:
                for li, ri in zip_longest(lfd, rfd, fillvalue=""):
                    if li.rstrip() != ri.rstrip():
                        self.lineending = False
                        break
                else:
                    # we went through every line, and they were all
                    # equal when stripped of their trailing whitespace
                    self.lineending = True

        return super(DistTextChange, self).check()


    def is_ignored(self, options):
        return (DistContentChange.is_ignored(self, options) or
                (self.lineending and options.ignore_trailing_whitespace))

    def collect_impl(self):
        with self.open_left(mode="rt") as lfd, \
             self.open_right(mode="rt") as rfd:

            left = lfd.read()
            right = rfd.read()
            if left != right:
                yield GenericChange(left, right)


class DistManifestChange(DistContentChange):
    """
    A MANIFEST.MF file found in the directory structure of the
    distribution
    """

    label = "Distributed Manifest"


    def collect_impl(self):
        if self.is_change():
            left_m = Manifest()
            left_m.parse_file(self.left_fn())
            right_m = Manifest()
            right_m.parse_file(self.right_fn())

            yield ManifestChange(left_m, right_m)


class DistJarChange(DistContentChange):

    label = "Distributed JAR"


    def collect_impl(self):
        if self.is_change():
            yield JarChange(self.left_fn(), self.right_fn())


class DistJarReport(DistJarChange):

    report_name = "JarReport"


    def __init__(self, ldata, rdata, entry, reporter):
        super(DistJarReport, self).__init__(ldata, rdata, entry, True)
        self.reporter = reporter


    def collect_impl(self):
        if self.is_change():
            yield JarReport(self.left_fn(), self.right_fn(), self.reporter)


class DistJarAdded(DistContentAdded):

    label = "Distributed JAR Added"


class DistJarRemoved(DistContentRemoved):

    label = "Distributed JAR Removed"


class DistClassChange(DistContentChange):

    label = "Distributed Java Class"


    def collect_impl(self):
        if self.is_change():
            linfo = unpack_classfile(self.left_fn())
            rinfo = unpack_classfile(self.right_fn())

            yield JavaClassChange(linfo, rinfo)


class DistClassReport(DistClassChange):

    report_name = "JavaClassReport"


    def __init__(self, l, r, entry, reporter):
        super(DistClassReport, self).__init__(l, r, entry, True)
        self.reporter = reporter


    def collect_impl(self):
        if self.is_change():
            linfo = unpack_classfile(self.left_fn())
            rinfo = unpack_classfile(self.right_fn())

            yield JavaClassReport(linfo, rinfo, self.reporter)


class DistClassAdded(DistContentAdded):

    label = "Distributed Java Class Added"


class DistClassRemoved(DistContentRemoved):

    label = "Distributed Java Class Removed"


class DistChange(SuperChange):
    """
    Top-level change for comparing two distributions
    """

    label = "Distribution"


    def __init__(self, left, right, shallow=False):
        super(DistChange, self).__init__(left, right)
        self.shallow = shallow


    def get_description(self):
        changed = "changed" if self.is_change() else "unchanged"
        return "%s %s from %s to %s" % \
            (self.label, changed, self.ldata, self.rdata)


    @yield_sorted_by_type(DistClassAdded,
                          DistClassRemoved,
                          DistClassChange,
                          DistJarAdded,
                          DistJarRemoved,
                          DistJarChange,
                          DistTextChange,
                          DistManifestChange,
                          DistContentAdded,
                          DistContentRemoved,
                          DistContentChange)
    def collect_impl(self):
        """
        emits change instances based on the delta of the two distribution
        directories
        """

        ld = self.ldata
        rd = self.rdata
        deep = not self.shallow

        for event, entry in compare(ld, rd):
            if deep and fnmatches(entry, *JAR_PATTERNS):
                if event == LEFT:
                    yield DistJarRemoved(ld, rd, entry)
                elif event == RIGHT:
                    yield DistJarAdded(ld, rd, entry)
                elif event == DIFF:
                    yield DistJarChange(ld, rd, entry, True)
                elif event == SAME:
                    yield DistJarChange(ld, rd, entry, False)

            elif deep and fnmatches(entry, "*.class"):
                if event == LEFT:
                    yield DistClassRemoved(ld, rd, entry)
                elif event == RIGHT:
                    yield DistClassAdded(ld, rd, entry)
                elif event == DIFF:
                    yield DistClassChange(ld, rd, entry, True)
                elif event == SAME:
                    yield DistClassChange(ld, rd, entry, False)

            elif deep and fnmatches(entry, *TEXT_PATTERNS):
                if event == LEFT:
                    yield DistContentRemoved(ld, rd, entry)
                elif event == RIGHT:
                    yield DistContentAdded(ld, rd, entry)
                elif event == DIFF:
                    yield DistTextChange(ld, rd, entry)
                elif event == SAME:
                    yield DistTextChange(ld, rd, entry, False)

            elif deep and fnmatches(entry, "*/MANIFEST.MF"):
                if event == LEFT:
                    yield DistContentRemoved(ld, rd, entry)
                elif event == RIGHT:
                    yield DistContentAdded(ld, rd, entry)
                elif event == DIFF:
                    yield DistManifestChange(ld, rd, entry, True)
                elif event == SAME:
                    yield DistManifestChange(ld, rd, entry, False)

            else:
                if event == LEFT:
                    yield DistContentRemoved(ld, rd, entry)
                elif event == RIGHT:
                    yield DistContentAdded(ld, rd, entry)
                elif event == DIFF:
                    yield DistContentChange(ld, rd, entry, True)
                elif event == SAME:
                    yield DistContentChange(ld, rd, entry, False)


class DistReport(DistChange):
    """
    This class has side-effects. Running the check method with the
    reportdir option set to True will cause the deep checks to be
    written to file in that directory
    """

    report_name = "DistReport"


    def __init__(self, l, r, reporter):
        self.reporter = reporter
        options = reporter.options
        shallow = getattr(options, "shallow", False)
        DistChange.__init__(self, l, r, shallow)


    def collect_impl(self):
        """
        overrides DistJarChange and DistClassChange from the underlying
        DistChange with DistJarReport and DistClassReport instances
        """

        for c in DistChange.collect_impl(self):
            if isinstance(c, DistJarChange):
                if c.is_change():
                    ln = DistJarReport.report_name
                    nr = self.reporter.subreporter(c.entry, ln)
                    c = DistJarReport(c.ldata, c.rdata, c.entry, nr)
            elif isinstance(c, DistClassChange):
                if c.is_change():
                    ln = DistClassReport.report_name
                    nr = self.reporter.subreporter(c.entry, ln)
                    c = DistClassReport(c.ldata, c.rdata, c.entry, nr)
            yield c


    def mp_check_impl(self, process_count):
        """
        a multiprocessing-enabled check implementation. Will create up to
        process_count helper processes and use them to perform the
        DistJarReport and DistClassReport actions.
        """

        from multiprocessing import Process, Queue

        options = self.reporter.options

        # this is the function that will be run in a separate process,
        # which will handle the tasks queue and feed into the results
        # queue
        func = _mp_run_check

        # normally this would happen lazily, but since we'll have
        # multiple processes all running reports at the same time, we
        # need to make sure the setup is done before-hand. This is
        # hackish, but in particular this keeps the HTML reports from
        # trying to perform the default data copy over and over.
        self.reporter.setup()

        # enqueue the sub-reports for multi-processing. Other types of
        # changes can happen sync.
        changes = list(self.collect_impl())

        task_count = 0
        tasks = Queue()
        results = Queue()

        try:
            # as soon as we start using the tasks queue, we need to be
            # catching the KeyboardInterrupt event so that we can
            # drain the queue and lets its underlying thread terminate
            # happily.

            # TODO: is there a better way to handle this shutdown
            # gracefully?

            # feed any sub-reports to the tasks queue
            for index in range(0, len(changes)):
                change = changes[index]
                if isinstance(change, (DistJarReport, DistClassReport)):
                    changes[index] = None
                    tasks.put((index, change))
                    task_count += 1

            # infrequent edge case, but don't bother starting more
            # helpers than we'll ever use
            process_count = min(process_count, task_count)

            # start the number of helper processes, and make sure
            # there are that many stop sentinels at the end of the
            # tasks queue
            for _i in range(0, process_count):
                tasks.put(None)
                process = Process(target=func, args=(tasks, results, options))
                process.daemon = False
                process.start()

            # while the helpers are running, perform our checks
            for change in changes:
                if change:
                    change.check()

            # get all of the results and feed them back into our change
            for _i in range(0, task_count):
                index, change = results.get()
                changes[index] = change

        except KeyboardInterrupt:
            # drain the tasks queue so it will exit gracefully
            for _change in iter(tasks.get, None):
                pass
            raise

        # complete the check by setting our internal collection of
        # child changes and returning our overall status
        c = False
        for change in changes:
            c = c or change.is_change()
        self.changes = changes
        return c, None


    def check_impl(self):
        options = self.reporter.options

        # if we're configured to use multiple processes, the work happens
        # in mp_check_impl instead
        forks = getattr(options, "processes", 0)
        if forks:
            return self.mp_check_impl(forks)

        changes = list()

        c = False
        for change in self.collect_impl():
            change.check()
            c = c or change.is_change()

            if isinstance(change, (DistJarReport, DistClassReport)):
                # the child report has run, we only need to keep the
                # squashed overview
                changes.append(squash(change, options=options))
                change.clear()
            else:
                changes.append(change)

        self.changes = changes
        return c, None


    def check(self):
        # do the actual checking
        DistChange.check(self)

        # write to file
        self.reporter.run(self)


def _mp_run_check(tasks, results, options):
    """
    a helper function for multiprocessing with DistReport.
    """

    try:
        for index, change in iter(tasks.get, None):
            # this is the part that takes up all of our time and
            # produces side-effects like writing out files for all of
            # the report formats.
            change.check()

            # rather than serializing the completed change (which
            # could be rather large now that it's been realized), we
            # send back only what we want, which is the squashed
            # overview, and throw away the used bits.
            squashed = squash(change, options=options)
            change.clear()

            results.put((index, squashed))

    except KeyboardInterrupt:
        # prevent a billion lines of backtrace from hitting the user
        # in the face
        return


# ---- Begin distdiff CLI ----
#


def cli_dist_diff(options, left, right):
    from .report import quick_report, Reporter
    from .report import JSONReportFormat, TextReportFormat

    reports = getattr(options, "reports", tuple())
    if reports:
        rdir = options.report_dir or "./"

        rpt = Reporter(rdir, "DistReport", options)
        rpt.add_formats_by_name(reports)

        delta = DistReport(left, right, rpt)

    else:
        delta = DistChange(left, right, options.shallow)

    delta.check()

    if not options.silent:
        if options.json:
            quick_report(JSONReportFormat, delta, options)
        else:
            quick_report(TextReportFormat, delta, options)

    if (not delta.is_change()) or delta.is_ignored(options):
        return 0
    else:
        return 1


def cli(options):
    left, right = options.dist
    return cli_dist_diff(options, left, right)


def add_distdiff_optgroup(parser):
    """
    Option group relating to the use of a DistChange or DistReport
    """

    # for the --processes default
    cpus = cpu_count()

    og = parser.add_argument_group("Distribution Checking Options")

    og.add_argument("--processes", type=int, default=cpus,
                    help="Number of child processes to spawn to handle"
                    " sub-reports. Set to 0 to disable multi-processing."
                    " Defaults to the number of CPUs (%r)" % cpus)

    og.add_argument("--shallow", action="store_true", default=False,
                    help="Check only that the files of this dist have"
                    "changed, do not infer the meaning")

    og.add_argument("--ignore-filenames", action="append", default=[],
                    help="file glob to ignore. Can be specified multiple"
                    " times")

    og.add_argument("--ignore-trailing-whitespace",
                    action="store_true", default=False,
                    help="ignore trailing whitespace when comparing text"
                    " files")


def create_optparser(progname=None):
    """
    an OptionParser instance filled with options and groups
    appropriate for use with the distdiff command
    """
    from . import report

    parser = ArgumentParser(prog=progname)
    parser.add_argument("dist", nargs=2,
                        help="distributions to compare")

    add_general_optgroup(parser)
    add_distdiff_optgroup(parser)
    add_jardiff_optgroup(parser)
    add_classdiff_optgroup(parser)

    report.add_general_report_optgroup(parser)
    report.add_json_report_optgroup(parser)
    report.add_html_report_optgroup(parser)

    return parser


def default_distdiff_options(updates=None):
    """
    generate an options object with the appropriate default values in
    place for API usage of distdiff features. overrides is an optional
    dictionary which will be used to update fields on the options
    object.
    """

    parser = create_optparser()
    options = parser.parse_args(list())

    if updates:
        # pylint: disable=W0212
        options._update_careful(updates)

    return options


def main(args=sys.argv):
    """
    entry point for the distdiff command-line utility
    """

    parser = create_optparser(args[0])
    return cli(parser.parse_args(args[1:]))


#
# The end.



#! /usr/bin/env python

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Python Javatools

author: Christopher O'Brien  <obriencj@gmail.com>
license: LGPL v.3
"""


from setuptools import setup


PYTHON_SUPPORTED_VERSIONS = (
    ">=2.7",
    "!=3.0.*", "!=3.1.*", "!=3.2.*", "!=3.3.*", "!=3.4.*",
    "<4",
)


def delayed_cheetah_build_py_cmd(*args, **kwds):
    # only import cheetah_build_py_cmd immediately before
    # instantiating it.
    from javatools.cheetah.setuptools import cheetah_build_py_cmd
    return cheetah_build_py_cmd(*args, **kwds)


setup(name = "javatools",
      version = "1.5.0",

      packages = [
          "javatools",
          "javatools.cheetah",
      ],

      package_data = {
          "javatools.cheetah": [
              "*.tmpl",
              "data/*.css",
              "data/*.js",
              "data/*.png",
          ],
      },

      entry_points = {
          "console_scripts": [
              'classdiff=javatools.classdiff:main',
              'classinfo=javatools.classinfo:main',
              'distdiff=javatools.distdiff:main',
              'distinfo=javatools.distinfo:main',
              'jardiff=javatools.jardiff:main',
              'jarinfo=javatools.jarinfo:main',
              'jarutil=javatools.jarutil:main',
              'manifest=javatools.manifest:main',
          ],
      },

      test_suite = "tests",

      # PyPI information
      author = "Christopher O'Brien",
      author_email = "obriencj@gmail.com",
      url = "https://github.com/obriencj/python-javatools",
      license = "GNU Lesser General Public License v.3",

      description = "Tools for finding meaningful deltas in Java"
      " class files and JARs",

      provides = [
          "javatools",
      ],

      install_requires = [
          "Cheetah3",
          "M2Crypto >= 0.26.0",
          "six",
      ],

      setup_requires = [
          "Cheetah3",
          "six",
      ],

      python_requires = ", ".join(PYTHON_SUPPORTED_VERSIONS),

      classifiers = [
          "Development Status :: 5 - Production/Stable",
          "Environment :: Console",
          "Intended Audience :: Developers",
          "Intended Audience :: Information Technology",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          # "Programming Language :: Python :: 3.7",
          "Topic :: Software Development :: Disassemblers",
      ],

      # extras stuff
      cmdclass = {'build_py': delayed_cheetah_build_py_cmd, })


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Some abstraction of changes. Useful for the classdiff and jardiff
modules.

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""


from functools import wraps


__all__ = (
    "squash",
    "collect_by_typename", "collect_by_type",
    "iterate_by_type", "yield_sorted_by_type",
    "Change", "Addition", "Removal",
    "GenericChange", "SuperChange",
    "SquashedChange", "SquashedAddition", "SquashedRemoval", )


def collect_by_typename(obj_sequence, cache=None):
    """
    collects objects from obj_sequence and stores them into buckets by
    type name. cache is an optional dict into which we collect the
    results.
    """

    if cache is None:
        cache = {}

    for val in obj_sequence:
        key = type(val).__name__

        bucket = cache.get(key, None)
        if bucket is not None:
            bucket.append(val)
        else:
            cache[key] = [val]

    return cache


def collect_by_type(obj_sequence, cache=None):
    """
    collects objects from obj_sequence and stores them into buckets by
    type. cache is an optional dict into which we collect the results.
    """

    if cache is None:
        cache = {}

    for val in obj_sequence:
        key = type(val)

        bucket = cache.get(key, None)
        if bucket is not None:
            bucket.append(val)
        else:
            cache[key] = [val]

    return cache


def iterate_by_type(objs, typelist):
    """
    collects a sequence of objs into buckets by type, then re-emits
    objs from the buckets, sorting through the buckets in the order
    specified by typelist. Any objects of a type not specified in
    typelist will be emitted last in no guaranteed order (but still
    grouped by type).
    """

    cache = collect_by_type(objs)

    for t in typelist:
        for val in cache.pop(t, tuple()):
            yield val

    for tl in cache.values():
        for val in tl:
            yield val


def yield_sorted_by_type(*typelist):
    """
    a useful decorator for the collect_impl method of SuperChange
    subclasses. Caches the yielded changes, and re-emits them
    collected by their type. The order of the types can be specified
    by listing the types as arguments to this decorator. Unlisted
    types will be yielded last in no guaranteed order.

    Grouping happens by exact type match only. Inheritance is not
    taken into consideration for grouping.
    """

    def decorate(fun):
        @wraps(fun)
        def decorated(*args, **kwds):
            return iterate_by_type(fun(*args, **kwds), typelist)
        return decorated

    return decorate


class Change(object):
    """
    Base class for representing a specific change between two objects
    """

    label = "Change"


    def __init__(self, ldata, rdata):
        self.ldata = ldata
        self.rdata = rdata
        self.description = None
        self.changed = False
        self.entry = None


    def __del__(self):
        self.clear()


    def clear(self):
        self.ldata = None
        self.rdata = None
        self.description = None
        self.changed = False
        self.entry = None


    def check(self):
        pass


    def get_ldata(self):
        return self.ldata


    def get_rdata(self):
        return self.rdata


    def is_change(self):
        return self.changed


    def is_ignored(self, options):
        """
        is this change ignorable, given parameters on the options
        object.
        """

        return False


    def get_description(self):
        return self.description or \
            (self.label + (" changed" if self.is_change() else " unchanged"))


    def collect(self, force=False):
        return tuple()


    def simplify(self, options=None):
        """
        returns a dict describing a simple snapshot of this change, and
        its children if any.
        """

        simple = {
            "class": type(self).__name__,
            "is_change": self.is_change(),
            "description": self.get_description(),
            "label": self.label,
        }

        if options:
            simple["is_ignored"] = self.is_ignored(options)

        if isinstance(self, Addition):
            simple["is_addition"] = True

        if isinstance(self, Removal):
            simple["is_removal"] = True

        if self.entry:
            simple["entry"] = self.entry

        return simple


class Removal(Change):
    """
    A type of change indicating that something was removed
    """

    label = "Removal"


    def is_change(self):
        return True


class Addition(Change):
    """
    A type of change indicating that something was added
    """

    label = "Addition"


    def is_change(self):
        return True


class GenericChange(Change):
    """
    A generalized test for a single change on two objects: a left and
    a right. Subclasses should override the label and the check_impl
    method at a minimum.
    """

    label = "Generic Change"


    def fn_data(self, side_data):
        """
        Get the data to be used in fn_differ from side_data. By default,
        this method is the identity
        """

        return side_data


    def fn_pretty(self, side_data):
        """
        override to provide a way to show the pretty version of the left
        or right data. Defaults to fn_data
        """

        return self.fn_data(side_data)


    def fn_pretty_desc(self, side_data):
        """
        override to provide a way to describe the data left or right
        data. Defaults to fn_pretty
        """

        return self.fn_pretty(side_data)


    def fn_differ(self, left_data, right_data):
        """
        override to provide the check for whether get_ldata() and
        get_rdata() differ. defaults to an inequality (!=) check
        """

        return left_data != right_data


    def get_ldata(self):
        """
        returns fn_data of ldata
        """

        return self.fn_data(self.ldata)


    def get_rdata(self):
        """
        returns fn_data of rdata
        """

        return self.fn_data(self.rdata)


    def pretty_ldata(self):
        """
        returns fn_pretty of ldata (NOT the fn_pretty of get_ldata)
        """

        return self.fn_pretty(self.ldata)


    def pretty_rdata(self):
        """
        returns fn_pretty of rdata (NOT the fn_pretty of get_rdata)
        """

        return self.fn_pretty(self.rdata)


    def pretty_ldata_desc(self):
        """
        returns fn_pretty_desc of ldata (NOT the fn_pretty_desc of
        get_ldata)
        """

        return self.fn_pretty_desc(self.ldata)


    def pretty_rdata_desc(self):
        """
        returns fn_pretty_desc of rdata (NOT the fn_pretty_desc of
        get_rdata)
        """

        return self.fn_pretty_desc(self.rdata)


    def check_impl(self):
        """
        returns a tuple of (is_change,description) which are then stored
        in self.changed and self.description

        The default implementation will get the data from the left and
        right sides by calling self.fn_data, then compare them via
        self.fn_differ. If they do differ, a message will be
        constructed using self.fn_pretty to create human-readable
        versions of the data that changed.
        """

        if self.fn_differ(self.get_ldata(), self.get_rdata()):

            left = self.pretty_ldata_desc()
            right = self.pretty_rdata_desc()
            msg = "%s changed: %s to %s" % (self.label, left, right)

            return True, msg

        else:
            return False, None


    def check(self):
        """
        if necessary, override check_impl to change the behaviour of
        subclasses of GenericChange.
        """

        self.changed, self.description = self.check_impl()


    def simplify(self, options=None):
        """
        provide a simple representation of this change as a dictionary
        """

        # TODO: we might want to get rid of this method and just move
        # it into the JSONEncoder in report.py

        simple = super(GenericChange, self).simplify(options)

        ld = self.pretty_ldata()
        if ld is not None:
            simple["old_data"] = ld

        rd = self.pretty_rdata()
        if rd is not None:
            simple["new_data"] = rd

        return simple


class SuperChange(GenericChange):
    """
    A collection of changes.

    For simplest use, override the change_types class field with a
    list of Change subclasses. When the default collect_impl is called
    from collect, an instance of each type will be created with the
    same left and right data as the SuperChange instance was created
    with. The check_impl (called from check) will iterate over the
    instances and call their check method in-turn.

    An instance of SuperChange is considered unchanged if all of its
    sub-changes are also unchanged (or if there were no sub-changes).

    An instance of SuperChange is considered ignored if it was a
    change and all of its changed children were also ignored.
    """

    label = "Super Change"


    # override with change classes
    change_types = tuple()


    def __init__(self, ldata, rdata):
        super(SuperChange, self).__init__(ldata, rdata)
        self.changes = tuple()


    def fn_pretty(self, c):
        return None


    def clear(self):
        """
        clears all child changes and drops the reference to them
        """

        super(SuperChange, self).clear()

        for c in self.changes:
            c.clear()
        self.changes = tuple()


    def collect_impl(self):
        """
        instantiates each of the entries in in the overriden change_types
        field with the left and right data
        """

        ldata = self.get_ldata()
        rdata = self.get_rdata()

        for change_type in self.change_types:
            yield change_type(ldata, rdata)


    def collect(self, force=False):
        """
        calls collect_impl and stores the results as the child changes of
        this super-change. Returns a tuple of the data generated from
        collect_impl. Caches the result rather than re-computing each
        time, unless force is True
        """

        if force or not self.changes:
            self.changes = tuple(self.collect_impl())
        return self.changes


    def check_impl(self):
        """
        sets self.changes to the result of self.changes_impl, then if any
        member of those checks shows as a change, will return
        True,None
        """

        c = False
        for change in self.collect():
            change.check()
            c = c or change.is_change()
        return c, None


    def is_ignored(self, options):
        """
        If we have changed children and all the children which are changes
        are ignored, then we are ignored. Otherwise, we are not
        ignored
        """

        if not self.is_change():
            return False

        changes = self.collect()
        if not changes:
            return False

        for change in changes:
            if change.is_change() and not change.is_ignored(options):
                return False
        return True


    def simplify(self, options=None):
        """
        generate a simple dict representing this change data, and
        collecting all of the sub-change instances (which are NOT
        immediately simplified themselves)
        """

        data = super(SuperChange, self).simplify(options)

        show_ignored = False
        show_unchanged = False

        if options:
            show_ignored = getattr(options, "show_ignored", show_ignored)
            show_unchanged = getattr(options, "show_unchanged", show_unchanged)

        # build a list of sub-changes honoring show-ignored and
        # show-unchanged

        subs = list()
        for s in self.collect():
            if s.is_change():
                if show_ignored or not s.is_ignored(options):
                    subs.append(s)
            elif show_unchanged:
                subs.append(s)

        data["children"] = subs
        return data


    def squash_children(self, options):
        """
        reduces the memory footprint of this super-change by converting
        all child changes into squashed changes
        """

        oldsubs = self.collect()
        self.changes = tuple(squash(c, options=options) for c in oldsubs)

        for change in oldsubs:
            change.clear()


class SquashedChange(Change):
    """
    For when you want to keep just the overall data from a change,
    including whether it was ignored, but want to discard the more
    in-depth information.
    """

    label = "SquashedChange"


    def __init__(self, change, is_ignored=False):
        super(SquashedChange, self).__init__(None, None)

        self.label = change.label
        self.description = change.get_description()
        self.changed = change.is_change()
        self.ignored = is_ignored
        self.origclass = type(change)
        self.entry = getattr(change, "entry", None)


    def is_ignored(self, options):
        return self.ignored


    def is_change(self):
        return self.changed


    def simplify(self, options=None):
        simple = super(SquashedChange, self).simplify(options)
        simple["original_class"] = self.origclass.__name__
        return simple


    def clear(self):
        pass


class SquashedRemoval(SquashedChange, Removal):
    """
    Squashed change indicating something was removed
    """

    label = "SquashedRemoval"


class SquashedAddition(SquashedChange, Addition):
    """
    Squashed change indicating something was added
    """

    label = "SquashedAddition"


def squash(change, is_ignored=False, options=None):
    """
    squashes the in-depth information of a change to a simplified (and
    less memory-intensive) form
    """

    if options:
        is_ignored = change.is_ignored(options)

    if isinstance(change, Removal):
        result = SquashedRemoval(change, is_ignored)
    elif isinstance(change, Addition):
        result = SquashedAddition(change, is_ignored)
    else:
        result = SquashedChange(change, is_ignored)

    return result


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.



"""
Utility script and module for discovering information about a
distribution of mixed class files and JARs.

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""


from __future__ import print_function


import sys

from json import dump
from argparse import ArgumentParser
from os.path import isdir, join
from shutil import rmtree
from tempfile import mkdtemp

from . import unpack_classfile
from .jarinfo import JarInfo, JAR_PATTERNS, REQ_BY_CLASS, PROV_BY_CLASS
from .dirutils import fnmatches
from .ziputils import open_zip


DIST_JAR = "jar"
DIST_CLASS = "class"

REQ_BY_JAR = "jar.requires"
PROV_BY_JAR = "jar.provides"


class DistInfo(object):


    def __init__(self, base_path):
        self.base_path = base_path

        # a pair of strings useful for later reporting. Non-mandatory
        self.product = None
        self.version = None

        # if the dist is a zip, we'll explode it into tmpdir
        self.tmpdir = None

        self._contents = None
        self._requires = None
        self._provides = None


    def __del__(self):
        self.close()


    def _working_path(self):
        if self.tmpdir:
            return self.tmpdir

        elif isdir(self.base_path):
            return self.base_path

        else:
            self.tmpdir = mkdtemp()
            with open_zip(self.base_path, "r") as zf:
                zf.extractall(path=self.tmpdir)
            return self.tmpdir


    def _collect_requires_provides(self):
        req = {}
        prov = {}

        p = set()

        for entry in self.get_jars():
            ji = self.get_jarinfo(entry)
            for sym, data in ji.get_requires().items():
                req.setdefault(sym, []).append((REQ_BY_JAR, entry, data))
            for sym, data in ji.get_provides().items():
                prov.setdefault(sym, []).append((PROV_BY_JAR, entry, data))
                p.add(sym)
            ji.close()

        for entry in self.get_classes():
            ci = self.get_classinfo(entry)
            for sym in ci.get_requires():
                req.setdefault(sym, []).append((REQ_BY_CLASS, entry))
            for sym in ci.get_provides(private=False):
                prov.setdefault(sym, []).append((PROV_BY_CLASS, entry))
            for sym in ci.get_provides(private=True):
                p.add(sym)

        req = dict((k, v) for k, v in req.items() if k not in p)

        self._requires = req
        self._provides = prov


    def get_requires(self, ignored=tuple()):
        """ a map of requirements to what requires it. ignored is an
        optional list of globbed patterns indicating packages,
        classes, etc that shouldn't be included in the provides map"""

        if self._requires is None:
            self._collect_requires_provides()

        d = self._requires
        if ignored:
            d = dict((k, v) for k, v in d.items()
                     if not fnmatches(k, *ignored))
        return d


    def get_provides(self, ignored=tuple()):
        """ a map of provided classes and class members, and what
        provides them. ignored is an optional list of globbed patterns
        indicating packages, classes, etc that shouldn't be included
        in the provides map"""

        if self._provides is None:
            self._collect_requires_provides()

        d = self._provides
        if ignored:
            d = dict((k, v) for k, v in d.items()
                     if not fnmatches(k, *ignored))
        return d


    def get_jars(self):
        """ sequence of entry names found in this distribution """

        for entry in self.get_contents():
            if fnmatches(entry, *JAR_PATTERNS):
                yield entry


    def get_jarinfo(self, entry):
        return JarInfo(join(self.base_path, entry))


    def get_classes(self):
        """ sequence of entry names found in the distribution.  This
        is only the collection of class files directly in the dist, it
        does not include classes within JARs that are inthe dist."""

        for entry in self.get_contents():
            if fnmatches(entry, "*.class"):
                yield entry


    def get_classinfo(self, entry):
        return unpack_classfile(join(self.base_path, entry))


    def get_contents(self):
        if self._contents is None:
            self._contents = tuple(_collect_dist(self._working_path()))
        return self._contents


    def close(self):
        """ if this was a zip'd distribution, any introspection
        may have resulted in opening or creating temporary files.
        Call close in order to clean up. """

        if self.tmpdir:
            rmtree(self.tmpdir)
            self.tmpdir = None

        self._contents = None


def _collect_dist(pathn):
    from os.path import join, relpath
    from os import walk
    for r, _ds, fs in walk(pathn):
        for f in fs:
            yield relpath(join(r, f), pathn)


# --- CLI ---
#


def cli_dist_provides(options, info):
    print("distribution provides:")

    for provided in sorted(info.get_provides(options.api_ignore)):
        print(" ", provided)
    print()


def cli_dist_requires(options, info):
    print("distribution requires:")

    for required in sorted(info.get_requires(options.api_ignore)):
        print(" ", required)
    print()


def cli_distinfo(options, info):

    if options.dist_provides:
        cli_dist_provides(options, info)

    if options.dist_requires:
        cli_dist_requires(options, info)

    # TODO: simple things like listing JARs and non-JAR files


def cli_distinfo_json(options, info):
    data = {}

    if options.dist_provides:
        data["dist.provides"] = info.get_provides(options.api_ignore)

    if options.dist_requires:
        data["dist.requires"] = info.get_requires(options.api_ignore)

    dump(data, sys.stdout, sort_keys=True, indent=2)


def cli(options):
    # pylint: disable=W0613
    # parser unused

    pathn = options.dist
    info = DistInfo(pathn)

    if options.json:
        cli_distinfo_json(options, info)
    else:
        cli_distinfo(options, info)

    info.close()
    return 0


def add_distinfo_optgroup(parser):
    g = parser.add_argument_group("Distribution Info Options")

    g.add_argument("--dist-provides", dest="dist_provides",
                   action="store_true", default=False,
                   help="API provides information at the distribution level")

    g.add_argument("--dist-requires", dest="dist_requires",
                   action="store_true", default=False,
                   help="API requires information at the distribution level")


def create_optparser(progname):
    from .jarinfo import add_jarinfo_optgroup
    from .classinfo import add_classinfo_optgroup

    parser = ArgumentParser(prog=progname)
    parser.add_argument("dist", help="distribution to inspect")
    parser.add_argument("--json", dest="json", action="store_true",
                        help="output in JSON mode")

    add_distinfo_optgroup(parser)
    add_jarinfo_optgroup(parser)
    add_classinfo_optgroup(parser)

    return parser


def main(args=sys.argv):
    parser = create_optparser(args[0])
    return cli(parser.parse_args(args[1:]))


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Classes for representing changes as formatted text.

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""


from __future__ import print_function


import sys

from abc import ABCMeta, abstractmethod
from argparse import Action
from Cheetah.DummyTransaction import DummyTransaction
from functools import partial
from json import dump, JSONEncoder
from os.path import exists, join, relpath
from six import add_metaclass
from six.moves import range

from .dirutils import copydir, makedirsp


_BUFFERING = 2 ** 16


__all__ = (
    "Reporter", "ReportFormat",
    "quick_report", "add_general_report_optgroup",
    "JSONReportFormat", "add_json_report_optgroup",
    "TextReportFormat",
    "CheetahReportFormat", "add_html_report_optgroup", )


class Reporter(object):
    """
    Collects multiple report formats for use in presenting a change
    """


    def __init__(self, basedir, entry, options):
        self.basedir = basedir
        self.entry = entry
        self.options = options
        self.breadcrumbs = tuple()
        self.formats = set()

        # cache of instances from self.formats, created in setup, used
        # in run, removed in clear
        self._formats = None


    def get_relative_breadcrumbs(self):
        """
        get the breadcrumbs as relative to the basedir
        """

        basedir = self.basedir
        crumbs = self.breadcrumbs

        return [(relpath(b, basedir), e) for b, e in crumbs]


    def add_formats_by_name(self, rfmt_list):
        """
        adds formats by short label descriptors, such as 'txt', 'json', or
        'html'
        """

        for fmt in rfmt_list:
            if fmt == "json":
                self.add_report_format(JSONReportFormat)
            elif fmt in ("txt", "text"):
                self.add_report_format(TextReportFormat)
            elif fmt in ("htm", "html"):
                self.add_report_format(CheetahReportFormat)


    def add_report_format(self, report_format):
        """
        Add an output format to this reporter. report_format should be a
        ReportFormat subtype. It will be instantiated when the
        reporter is run.
        """

        self.formats.add(report_format)


    def subreporter(self, subpath, entry):
        """
        create a reporter for a sub-report, with updated breadcrumbs and
        the same output formats
        """

        newbase = join(self.basedir, subpath)
        r = Reporter(newbase, entry, self.options)

        crumbs = list(self.breadcrumbs)
        crumbs.append((self.basedir, self.entry))
        r.breadcrumbs = crumbs

        r.formats = set(self.formats)
        return r


    def setup(self):
        """
        instantiates all report formats that have been added to this
        reporter, and calls their setup methods.
        """

        if self._formats:
            # setup has been run already.
            return

        basedir = self.basedir
        options = self.options
        crumbs = self.get_relative_breadcrumbs()

        fmts = list()
        for fmt_class in self.formats:
            fmt = fmt_class(basedir, options, crumbs)
            fmt.setup()
            fmts.append(fmt)

        self._formats = fmts


    def run(self, change):
        """
        runs the report format instances in this reporter. Will call setup
        if it hasn't been called already
        """

        if self._formats is None:
            self.setup()

        entry = self.entry

        for fmt in self._formats:
            fmt.run(change, entry)

        self.clear()


    def clear(self):
        """
        calls clear on any report format instances created during setup
        and drops the cache
        """

        if self._formats:
            for fmt in self._formats:
                fmt.clear()
        self._formats = None


@add_metaclass(ABCMeta)
class ReportFormat(object):
    """
    Base class of a report format provider. Override to describe a
    concrete format type
    """

    extension = ".report"


    def __init__(self, basedir, options, breadcrumbs=tuple()):
        self.basedir = basedir
        self.options = options
        self.breadcrumbs = breadcrumbs


    @abstractmethod
    def run_impl(self, change, entry, out):
        """
        override to actually produce output
        """

        pass


    def run(self, change, entry, out=None):
        """
        setup for run, including creating an output file if needed. Calls
        run_impl when ready. If out and entry are both None,
        sys.stdout is used
        """

        if out:
            self.run_impl(change, entry, out)
            return None

        elif entry:
            basedir = self.basedir or "./"

            makedirsp(basedir)

            fn = join(basedir, entry + self.extension)
            with open(fn, "wt", _BUFFERING) as out:
                self.run_impl(change, entry, out)
            return fn

        else:
            self.run_impl(change, entry, sys.stdout)
            return None


    def setup(self):
        """
        override if the report format has behavior which can be done ahead
        of time, such as copying files or creating directories
        """

        pass


    def clear(self):
        """
        clear up the internal references of this format. Called by the
        parent reporter at the end of run
        """

        self.basedir = None
        self.options = None
        self.breadcrumbs = None


class _opt_cb_report(Action):

    """
    callback for the --report option in general_report_optgroup
    """
    def __call__(self, parser, options, values, option_string=None):

        if not hasattr(options, "reports"):
            options.reports = list()

        if "," in values:
            options.reports.extend(v for v in values.split(",") if v)
        else:
            options.reports.append(values)


def add_general_report_optgroup(parser):
    """
    General Reporting Options
    """

    g = parser.add_argument_group("Reporting Options")

    g.add_argument("--report-dir", action="store", default=None)

    g.add_argument("--report", action=_opt_cb_report,
                   help="comma-separated list of report formats")


class JSONReportFormat(ReportFormat):
    """
    renders a Change and all of its children to a JSON object. Can use
    options from the jon_report_optgroup option group
    """

    extension = ".json"


    def run_impl(self, change, entry, out):
        options = self.options
        indent = getattr(options, "json_indent", 2)

        data = {
            "runtime_options": options.__dict__,
            "report": change,
        }

        # not what they expected, but it works
        cls = partial(JSONChangeEncoder, options)

        try:
            dump(data, out, sort_keys=True, indent=indent, cls=cls)

        except TypeError:
            # XXX for debugging. Otherwise the wrapping try isn't necessary
            print(data)
            raise


def add_json_report_optgroup(parser):
    """
    Option group for the JSON report format
    """

    g = parser.add_argument_group("JSON Report Options")

    g.add_argument("--json-indent", action="store", default=2, type=int)


class JSONChangeEncoder(JSONEncoder):
    """
    A specialty JSONEncoder which knows how to represent Change
    instances (or anything with a simplify method), and sequences (by
    rendering them into tuples)
    """


    def __init__(self, options, *a, **k):
        JSONEncoder.__init__(self, *a, **k)
        self.options = options


    def default(self, o):
        # pylint: disable=E0202
        # JSONEncoder.default confuses pylint

        # if there is a simplify method, call it to convert the object
        # into a simplified dict
        if hasattr(o, "simplify"):
            return o.simplify(self.options)

        # handle sequences sanely
        try:
            i = iter(o)
        except TypeError:
            return JSONEncoder.default(self, o)
        else:
            return tuple(i)


class TextReportFormat(ReportFormat):
    """
    Renders the change as indented text
    """

    extension = ".text"


    def run_impl(self, change, entry, out):
        options = self.options
        _indent_change(change, out, options, 0)


def _indent_change(change, out, options, indent):
    """
    recursive function to print indented change descriptions
    """

    show_unchanged = getattr(options, "show_unchanged", False)
    show_ignored = getattr(options, "show_ignored", False)

    show = False
    desc = change.get_description()

    if change.is_change():
        if change.is_ignored(options):
            if show_ignored:
                show = True
                _indent(out, indent, desc, " [IGNORED]")
        else:
            show = True
            _indent(out, indent, desc)
    elif show_unchanged:
        show = True
        _indent(out, indent, desc)

    if show:
        indent += 1
        for sub in change.collect():
            _indent_change(sub, out, options, indent)


def _indent(stream, indent, *msgs):

    """ write a message to a text stream, with indentation. Also ensures that
    the output encoding of the messages is safe for writing. """

    for x in range(0, indent):
        stream.write("  ")
    for x in msgs:
        # Any nicer way? In Py2 x can be 'str' or 'unicode'.
        stream.write(x.encode("ascii", "backslashreplace").decode("ascii"))
    stream.write("\n")


class CheetahReportFormat(ReportFormat):
    """
    HTML output for a Change
    """

    extension = ".html"


    def _relative(self, uri):
        """
        if uri is relative, re-relate it to our basedir
        """

        if uri.startswith("http:") or \
           uri.startswith("https:") or \
           uri.startswith("file:") or \
           uri.startswith("/"):
            return uri

        elif exists(uri):
            return relpath(uri, self.basedir)

        else:
            return uri


    def _relative_uris(self, uri_list):
        """
        if uris in list are relative, re-relate them to our basedir
        """

        return [u for u in (self._relative(uri) for uri in uri_list) if u]


    def setup(self):
        """
        copies default stylesheets and javascript files if necessary, and
        appends them to the options
        """

        from javatools import cheetah

        options = self.options
        datadir = getattr(options, "html_copy_data", None)

        if getattr(options, "html_data_copied", False) or not datadir:
            # either already run by a parent report, or not supposed
            # to run at all.
            return

        # this is where we've installed the default media
        datasrc = join(cheetah.__path__[0], "data")

        # record the .js and .css content we copy
        javascripts = list()
        stylesheets = list()

        # copy the contents of our data source to datadir
        for _orig, copied in copydir(datasrc, datadir):
            if copied.endswith(".js"):
                javascripts.append(copied)
            elif copied.endswith(".css"):
                stylesheets.append(copied)

        javascripts.extend(getattr(options, "html_javascripts", tuple()))
        stylesheets.extend(getattr(options, "html_stylesheets", tuple()))

        options.html_javascripts = javascripts
        options.html_stylesheets = stylesheets

        # keep from copying again
        options.html_data_copied = True


    def run_impl(self, change, entry, out):
        """
        sets up the report directory for an HTML report. Obtains the
        top-level Cheetah template that is appropriate for the change
        instance, and runs it.

        The cheetah templates are supplied the following values:
         * change - the Change instance to report on
         * entry - the string name of the entry for this report
         * options - the cli options object
         * breadcrumbs - list of backlinks
         * javascripts - list of .js links
         * stylesheets - list of .css links

        The cheetah templates are also given a render_change method
        which can be called on another Change instance to cause its
        template to be resolved and run in-line.
        """

        options = self.options

        # translate relative paths if necessary
        javascripts = self._relative_uris(options.html_javascripts)
        stylesheets = self._relative_uris(options.html_stylesheets)

        # map the class of the change to a template class
        template_class = resolve_cheetah_template(type(change))

        # instantiate and setup the template
        template = template_class()

        # create a transaction wrapping the output stream
        template.transaction = DummyTransaction()
        template.transaction.response(resp=out)

        # inject our values
        template.change = change
        template.entry = entry
        template.options = options
        template.breadcrumbs = self.breadcrumbs
        template.javascripts = javascripts
        template.stylesheets = stylesheets

        # this lets us call render_change from within the template on
        # a change instance to chain to another template (eg, for
        # sub-changes)
        template.render_change = lambda c: self.run_impl(c, entry, out)

        # execute the template, which will write its contents to the
        # transaction (and the underlying stream)
        template.respond()

        # clean up the template
        template.shutdown()


def _compose_cheetah_template_map(cache):
    """
    does the work of composing the cheetah template map into the given
    cache
    """

    from .cheetah import get_templates

    # pylint: disable=W0406
    # needed for introspection
    import javatools

    for template_type in get_templates():
        if "_" not in template_type.__name__:
            # we use the _ to denote package and class names. So any
            # template without a _ in the name isn't meant to be
            # matched to a change type.
            continue

        # get the package and change class names based off of the
        # template class name
        tn = template_type.__name__
        pn, cn = tn.split("_", 1)

        # get the package from javatools
        pk = getattr(javatools, pn, None)
        if pk is None:
            __import__("javatools." + pn)
            pk = getattr(javatools, pn, None)

        # get the class from the package
        cc = getattr(pk, cn, None)
        if cc is None:
            raise Exception("no change class for template %s" % tn)

        # associate a Change class with a Template class
        cache[cc] = template_type

    return cache


# pylint: disable=C0103
_template_cache = dict()


def cheetah_template_map(cache=None):
    """
    a map of change types to cheetah template types. Used in
    resolve_cheetah_template
    """

    if cache is None:
        cache = _template_cache

    return cache or _compose_cheetah_template_map(cache)


def resolve_cheetah_template(change_type):
    """
    return the appropriate cheetah template class for the given change
    type, using the method-resolution-order of the change type.
    """

    tm = cheetah_template_map()

    # follow the built-in MRO for a type to find the matching
    # cheetah template
    for t in change_type.mro():
        tmpl = tm.get(t)
        if tmpl:
            return tmpl

    # this should never happen, since we'll provide a
    # change_Change template, and all of the changes should be
    # inheriting from that type
    raise Exception("No template for class %s" % change_type.__name__)


def add_html_report_optgroup(parser):
    """
    Option group for the HTML report format
    """

    g = parser.add_argument_group("HTML Report Options")

    g.add_argument("--html-stylesheet", action="append",
                   dest="html_stylesheets", default=list())

    g.add_argument("--html-javascript", action="append",
                   dest="html_javascripts", default=list())

    g.add_argument("--html-copy-data", action="store", default=None,
                   help="Copy default resources to the given directory and"
                   " enable them in the template")


def quick_report(report_type, change, options):
    """
    writes a change report via report_type to options.output or
    sys.stdout
    """

    report = report_type(None, options)

    if options.output:
        with open(options.output, "w") as out:
            report.run(change, None, out)
    else:
        report.run(change, None, sys.stdout)


#
# The end.
